from __future__ import absolute_import, unicode_literals

__all__ = ["StandardDatabase", "AsyncDatabase", "BatchDatabase", "TransactionDatabase"]

from datetime import datetime

from .api import APIWrapper
from .aql import AQL
from .executor import (
    DefaultExecutor,
    AsyncExecutor,
    BatchExecutor,
    TransactionExecutor,
)
from .cluster import Cluster
from .collection import StandardCollection
from .exceptions import (
    AnalyzerCreateError,
    AnalyzerDeleteError,
    AnalyzerGetError,
    AnalyzerListError,
    AsyncJobClearError,
    AsyncJobListError,
    CollectionCreateError,
    CollectionDeleteError,
    CollectionListError,
    DatabaseDeleteError,
    DatabaseCreateError,
    DatabaseListError,
    DatabasePropertiesError,
    GraphListError,
    GraphCreateError,
    GraphDeleteError,
    PermissionListError,
    PermissionGetError,
    PermissionResetError,
    PermissionUpdateError,
    ServerEngineError,
    ServerDetailsError,
    ServerEchoError,
    JWTSecretListError,
    JWTSecretReloadError,
    ServerLogLevelError,
    ServerLogLevelSetError,
    ServerMetricsError,
    ServerReadLogError,
    ServerReloadRoutingError,
    ServerRequiredDBVersionError,
    ServerRoleError,
    ServerRunTestsError,
    ServerShutdownError,
    ServerStatisticsError,
    ServerStatusError,
    ServerTimeError,
    ServerTLSError,
    ServerTLSReloadError,
    ServerVersionError,
    TaskCreateError,
    TaskDeleteError,
    TaskGetError,
    TaskListError,
    TransactionExecuteError,
    UserCreateError,
    UserDeleteError,
    UserGetError,
    UserListError,
    UserReplaceError,
    UserUpdateError,
    ViewCreateError,
    ViewDeleteError,
    ViewGetError,
    ViewListError,
    ViewRenameError,
    ViewReplaceError,
    ViewUpdateError,
)
from .formatter import format_database, format_tls, format_view
from .foxx import Foxx
from .graph import Graph
from .pregel import Pregel
from .replication import Replication
from .request import Request
from .wal import WAL
from .utils import get_col_name


class Database(APIWrapper):
    """Base class for Database API wrappers.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param executor: API executor.
    :type executor: arango.executor.Executor
    """

    def __init__(self, connection, executor):
        super(Database, self).__init__(connection, executor)

    def __getitem__(self, name):
        """Return the collection API wrapper.

        :param name: Collection name.
        :type name: str | unicode
        :return: Collection API wrapper.
        :rtype: arango.collection.StandardCollection
        """
        return self.collection(name)

    def _get_col_by_doc(self, document):
        """Return the collection of the given document.

        :param document: Document ID or body with "_id" field.
        :type document: str | unicode | dict
        :return: Collection API wrapper.
        :rtype: arango.collection.StandardCollection
        :raise arango.exceptions.DocumentParseError: On malformed document.
        """
        return self.collection(get_col_name(document))

    @property
    def name(self):
        """Return database name.

        :return: Database name.
        :rtype: str | unicode
        """
        return self.db_name

    @property
    def aql(self):
        """Return AQL (ArangoDB Query Language) API wrapper.

        :return: AQL API wrapper.
        :rtype: arango.aql.AQL
        """
        return AQL(self._conn, self._executor)

    @property
    def wal(self):
        """Return WAL (Write-Ahead Log) API wrapper.

        :return: WAL API wrapper.
        :rtype: arango.wal.WAL
        """
        return WAL(self._conn, self._executor)

    @property
    def foxx(self):
        """Return Foxx API wrapper.

        :return: Foxx API wrapper.
        :rtype: arango.foxx.Foxx
        """
        return Foxx(self._conn, self._executor)

    @property
    def pregel(self):
        """Return Pregel API wrapper.

        :return: Pregel API wrapper.
        :rtype: arango.pregel.Pregel
        """
        return Pregel(self._conn, self._executor)

    @property
    def replication(self):
        """Return Replication API wrapper.

        :return: Replication API wrapper.
        :rtype: arango.replication.Replication
        """
        return Replication(self._conn, self._executor)

    @property
    def cluster(self):  # pragma: no cover
        """Return Cluster API wrapper.

        :return: Cluster API wrapper.
        :rtype: arango.cluster.Cluster
        """
        return Cluster(self._conn, self._executor)

    async def properties(self):
        """Return database properties.

        :return: Database properties.
        :rtype: dict
        :raise arango.exceptions.DatabasePropertiesError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/database/current",)

        def response_handler(resp):
            if not resp.is_success:
                raise DatabasePropertiesError(resp, request)
            return format_database(resp.body["result"])

        return await self._execute(request, response_handler)

    async def execute_transaction(
        self,
        command,
        params=None,
        read=None,
        write=None,
        sync=None,
        timeout=None,
        max_size=None,
        allow_implicit=None,
        intermediate_commit_count=None,
        intermediate_commit_size=None,
    ):
        """Execute raw Javascript command in transaction.

        :param command: Javascript command to execute.
        :type command: str | unicode
        :param read: Names of collections read during transaction. If parameter
            **allow_implicit** is set to True, any undeclared read collections
            are loaded lazily.
        :type read: [str | unicode]
        :param write: Names of collections written to during transaction.
            Transaction fails on undeclared write collections.
        :type write: [str | unicode]
        :param params: Optional parameters passed into the Javascript command.
        :type params: dict
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param timeout: Timeout for waiting on collection locks. If set to 0,
            ArangoDB server waits indefinitely. If not set, system default
            value is used.
        :type timeout: int
        :param max_size: Max transaction size limit in bytes. Applies only
            to RocksDB storage engine.
        :type max_size: int
        :param allow_implicit: If set to True, undeclared read collections are
            loaded lazily. If set to False, transaction fails on any undeclared
            collections.
        :type allow_implicit: bool
        :param intermediate_commit_count: Max number of operations after which
            an intermediate commit is performed automatically. Applies only to
            RocksDB storage engine.
        :type intermediate_commit_count: int
        :param intermediate_commit_size: Max size of operations in bytes after
            which an intermediate commit is performed automatically. Applies
            only to RocksDB storage engine.
        :type intermediate_commit_size: int
        :return: Return value of **command**.
        :rtype: str | unicode
        :raise arango.exceptions.TransactionExecuteError: If execution fails.
        """
        collections = {"allowImplicit": allow_implicit}
        if read is not None:
            collections["read"] = read
        if write is not None:
            collections["write"] = write

        data = {"action": command}
        if collections:
            data["collections"] = collections
        if params is not None:
            data["params"] = params
        if timeout is not None:
            data["lockTimeout"] = timeout
        if sync is not None:
            data["waitForSync"] = sync
        if max_size is not None:
            data["maxTransactionSize"] = max_size
        if intermediate_commit_count is not None:
            data["intermediateCommitCount"] = intermediate_commit_count
        if intermediate_commit_size is not None:
            data["intermediateCommitSize"] = intermediate_commit_size

        request = Request(method="post", endpoint="/_api/transaction", data=data)

        def response_handler(resp):
            if not resp.is_success:
                raise TransactionExecuteError(resp, request)
            return resp.body.get("result")

        return await self._execute(request, response_handler)

    async def version(self):
        """Return ArangoDB server version.

        :return: Server version.
        :rtype: str | unicode
        :raise arango.exceptions.ServerVersionError: If retrieval fails.
        """
        request = Request(
            method="get", endpoint="/_api/version", params={"details": False}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise ServerVersionError(resp, request)
            return resp.body["version"]

        return await self._execute(request, response_handler)

    async def details(self):
        """Return ArangoDB server details.

        :return: Server details.
        :rtype: dict
        :raise arango.exceptions.ServerDetailsError: If retrieval fails.
        """
        request = Request(
            method="get", endpoint="/_api/version", params={"details": True}
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body["details"]
            raise ServerDetailsError(resp, request)

        return await self._execute(request, response_handler)

    async def status(self):
        """Return ArangoDB server status.

        :return: Server status.
        :rtype: dict
        :raise arango.exceptions.ServerStatusError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_admin/status",)

        def response_handler(resp):
            if not resp.is_success:
                raise ServerStatusError(resp, request)

            body = resp.body or {}
            if "operationMode" in body:
                body["operation_mode"] = body.pop("operationMode")
            if "serverInfo" in body:
                info = body["serverInfo"]
                if "writeOpsEnabled" in info:
                    info["write_ops_enabled"] = info.pop("writeOpsEnabled")
                if "readOnly" in info:
                    info["read_only"] = info.pop("readOnly")
                body["server_info"] = body.pop("serverInfo")
            return body

        return await self._execute(request, response_handler)

    async def required_db_version(self):
        """Return required version of target database.

        :return: Required version of target database.
        :rtype: str | unicode
        :raise arango.exceptions.ServerRequiredDBVersionError: If retrieval
            fails.
        """
        request = Request(method="get", endpoint="/_admin/database/target-version")

        def response_handler(resp):
            if not resp.is_success:
                raise ServerRequiredDBVersionError(resp, request)
            return resp.body["version"]

        return await self._execute(request, response_handler)

    async def engine(self):
        """Return the database engine details.

        :return: Database engine details.
        :rtype: str | unicode
        :raise arango.exceptions.ServerEngineError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/engine")

        def response_handler(resp):
            if resp.is_success:
                return resp.body
            raise ServerEngineError(resp, request)

        return await self._execute(request, response_handler)

    async def statistics(self, description=False):
        """Return server statistics.

        :return: Server statistics.
        :rtype: dict
        :raise arango.exceptions.ServerStatisticsError: If retrieval fails.
        """
        if description:
            endpoint = "/_admin/statistics-description"
        else:
            endpoint = "/_admin/statistics"

        request = Request(method="get", endpoint=endpoint)

        def response_handler(resp):
            if resp.is_success:
                resp.body.pop("code")
                resp.body.pop("error")
                return resp.body
            raise ServerStatisticsError(resp, request)

        return await self._execute(request, response_handler)

    async def role(self):
        """Return server role.

        :return: Server role. Possible values are "SINGLE" (server which is not
            in a cluster), "COORDINATOR" (cluster coordinator), "PRIMARY",
            "SECONDARY", "AGENT" (Agency node in a cluster) or "UNDEFINED".
        :rtype: str | unicode
        :raise arango.exceptions.ServerRoleError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_admin/server/role")

        def response_handler(resp):
            if not resp.is_success:
                raise ServerRoleError(resp, request)
            return resp.body.get("role")

        return await self._execute(request, response_handler)

    async def time(self):
        """Return server system time.

        :return: Server system time.
        :rtype: datetime.datetime
        :raise arango.exceptions.ServerTimeError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_admin/time")

        def response_handler(resp):
            if not resp.is_success:
                raise ServerTimeError(resp, request)
            return datetime.fromtimestamp(resp.body["time"])

        return await self._execute(request, response_handler)

    async def echo(self):
        """Return details of the last request (e.g. headers, payload).

        :return: Details of the last request.
        :rtype: dict
        :raise arango.exceptions.ServerEchoError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_admin/echo")

        def response_handler(resp):
            if not resp.is_success:
                raise ServerEchoError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def shutdown(self):  # pragma: no cover
        """Initiate server shutdown sequence.

        :return: True if the server was shutdown successfully.
        :rtype: bool
        :raise arango.exceptions.ServerShutdownError: If shutdown fails.
        """
        request = Request(method="delete", endpoint="/_admin/shutdown")

        def response_handler(resp):
            if not resp.is_success:
                raise ServerShutdownError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def run_tests(self, tests):  # pragma: no cover
        """Run available unittests on the server.

        :param tests: List of files containing the test suites.
        :type tests: [str | unicode]
        :return: Test results.
        :rtype: dict
        :raise arango.exceptions.ServerRunTestsError: If execution fails.
        """
        request = Request(method="post", endpoint="/_admin/test", data={"tests": tests})

        def response_handler(resp):
            if not resp.is_success:
                raise ServerRunTestsError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def read_log(
        self,
        upto=None,
        level=None,
        start=None,
        size=None,
        offset=None,
        search=None,
        sort=None,
    ):
        """Read the global log from server.

        :param upto: Return the log entries up to the given level (mutually
            exclusive with parameter **level**). Allowed values are "fatal",
            "error", "warning", "info" (default) and "debug".
        :type upto: str | unicode | int
        :param level: Return the log entries of only the given level (mutually
            exclusive with **upto**). Allowed values are "fatal", "error",
            "warning", "info" (default) and "debug".
        :type level: str | unicode | int
        :param start: Return the log entries whose ID is greater or equal to
            the given value.
        :type start: int
        :param size: Restrict the size of the result to the given value. This
            can be used for pagination.
        :type size: int
        :param offset: Number of entries to skip (e.g. for pagination).
        :type offset: int
        :param search: Return only the log entries containing the given text.
        :type search: str | unicode
        :param sort: Sort the log entries according to the given fashion, which
            can be "sort" or "desc".
        :type sort: str | unicode
        :return: Server log entries.
        :rtype: dict
        :raise arango.exceptions.ServerReadLogError: If read fails.
        """
        params = dict()
        if upto is not None:
            params["upto"] = upto
        if level is not None:
            params["level"] = level
        if start is not None:
            params["start"] = start
        if size is not None:
            params["size"] = size
        if offset is not None:
            params["offset"] = offset
        if search is not None:
            params["search"] = search
        if sort is not None:
            params["sort"] = sort

        request = Request(method="get", endpoint="/_admin/log", params=params)

        def response_handler(resp):
            if not resp.is_success:
                raise ServerReadLogError(resp, request)
            if "totalAmount" in resp.body:
                resp.body["total_amount"] = resp.body.pop("totalAmount")
            return resp.body

        return await self._execute(request, response_handler)

    async def log_levels(self):
        """Return current logging levels.

        :return: Current logging levels.
        :rtype: dict
        """
        request = Request(method="get", endpoint="/_admin/log/level")

        def response_handler(resp):
            if not resp.is_success:
                raise ServerLogLevelError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def set_log_levels(self, **kwargs):
        """Set the logging levels.

        This method takes arbitrary keyword arguments where the keys are the
        logger names and the values are the logging levels. For example:

        .. code-block:: python

            arango.set_log_levels(
                agency='DEBUG',
                collector='INFO',
                threads='WARNING'
            )

        Keys that are not valid logger names are ignored.

        :return: New logging levels.
        :rtype: dict
        """
        request = Request(method="put", endpoint="/_admin/log/level", data=kwargs)

        def response_handler(resp):
            if not resp.is_success:
                raise ServerLogLevelSetError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def reload_routing(self):
        """Reload the routing information.

        :return: True if routing was reloaded successfully.
        :rtype: bool
        :raise arango.exceptions.ServerReloadRoutingError: If reload fails.
        """
        request = Request(method="post", endpoint="/_admin/routing/reload")

        def response_handler(resp):
            if not resp.is_success:
                raise ServerReloadRoutingError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def metrics(self):
        """Return server metrics in Prometheus format.

        :return: Server metrics in Prometheus format.
        :rtype: str
        """
        request = Request(method="get", endpoint="/_admin/metrics")

        def response_handler(resp):
            if not resp.is_success:
                raise ServerMetricsError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def jwt_secrets(self):  # pragma: no cover
        """Return information on currently loaded JWT secrets.

        :return: Information on currently loaded JWT secrets.
        :rtype: dict
        """
        request = Request(method="get", endpoint="/_admin/server/jwt")

        def response_handler(resp):
            if not resp.is_success:
                raise JWTSecretListError(resp, request)
            return resp.body["result"]

        return await self._execute(request, response_handler)

    async def reload_jwt_secrets(self):  # pragma: no cover
        """Hot-reload JWT secrets.

        Calling this without payload reloads JWT secrets from disk. Only files
        specified via arangod startup option ``--server.jwt-secret-keyfile`` or
        ``--server.jwt-secret-folder`` are used. It is not possible to change
        the location where files are loaded from without restarting the server.

        :return: Information on reloaded JWT secrets.
        :rtype: dict
        """
        request = Request(method="post", endpoint="/_admin/server/jwt")

        def response_handler(resp):
            if not resp.is_success:
                raise JWTSecretReloadError(resp, request)
            return resp.body["result"]

        return await self._execute(request, response_handler)

    async def tls(self):
        """Return TLS data (server key, client-auth CA).

        :return: TLS data.
        :rtype: dict
        """
        request = Request(method="get", endpoint="/_admin/server/tls")

        def response_handler(resp):
            if not resp.is_success:
                raise ServerTLSError(resp, request)
            return format_tls(resp.body["result"])

        return await self._execute(request, response_handler)

    async def reload_tls(self):
        """Reload TLS data (server key, client-auth CA).

        :return: New TLS data.
        :rtype: dict
        """
        request = Request(method="post", endpoint="/_admin/server/tls")

        def response_handler(resp):
            if not resp.is_success:
                raise ServerTLSReloadError(resp, request)
            return format_tls(resp.body["result"])

        return await self._execute(request, response_handler)

    #######################
    # Database Management #
    #######################

    async def databases(self):
        """Return the names all databases.

        :return: Database names.
        :rtype: [str | unicode]
        :raise arango.exceptions.DatabaseListError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/database")

        def response_handler(resp):
            if not resp.is_success:
                raise DatabaseListError(resp, request)
            return resp.body["result"]

        return await self._execute(request, response_handler)

    async def has_database(self, name):
        """Check if a database exists.

        :param name: Database name.
        :type name: str | unicode
        :return: True if database exists, False otherwise.
        :rtype: bool
        """
        return name in await self.databases()

    async def create_database(
        self,
        name,
        users=None,
        replication_factor=None,
        write_concern=None,
        sharding=None,
    ):
        """Create a new database.

        :param name: Database name.
        :type name: str | unicode
        :param users: List of users with access to the new database, where each
            user is a dictionary with fields "username", "password", "active"
            and "extra" (see below for example). If not set, only the admin and
            current user are granted access.
        :type users: [dict]
        :param replication_factor: Default replication factor for collections
            created in this database. Special values include "satellite" which
            replicates the collection to every DBServer, and 1 which disables
            replication. Used for clusters only.
        :type replication_factor: str | int
        :param write_concern: Default write concern for collections created in
            this database. Determines how many copies of each shard are
            required to be in sync on different DBServers. If there are less
            than these many copies in the cluster a shard will refuse to write.
            Writes to shards with enough up-to-date copies will succeed at the
            same time, however. Value of this parameter can not be larger than
            the value of **replication_factor**. Used for clusters only.
        :type write_concern: int
        :param sharding: Sharding method used for new collections in this
            database. Allowed values are: "", "flexible" and "single". The
            first two are equivalent. Used for clusters only.
        :type sharding: str
        :return: True if database was created successfully.
        :rtype: bool
        :raise arango.exceptions.DatabaseCreateError: If create fails.

        Here is an example entry for parameter **users**:

        .. code-block:: python

            {
                'username': 'john',
                'password': 'password',
                'active': True,
                'extra': {'Department': 'IT'}
            }
        """
        data = {"name": name}

        options = {}
        if replication_factor is not None:
            options["replicationFactor"] = replication_factor
        if write_concern is not None:
            options["writeConcern"] = write_concern
        if sharding is not None:
            options["sharding"] = sharding
        if options:
            data["options"] = options

        if users is not None:
            data["users"] = [
                {
                    "username": user["username"],
                    "passwd": user["password"],
                    "active": user.get("active", True),
                    "extra": user.get("extra", {}),
                }
                for user in users
            ]

        request = Request(method="post", endpoint="/_api/database", data=data)

        def response_handler(resp):
            if not resp.is_success:
                raise DatabaseCreateError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def delete_database(self, name, ignore_missing=False):
        """Delete the database.

        :param name: Database name.
        :type name: str | unicode
        :param ignore_missing: Do not raise an exception on missing database.
        :type ignore_missing: bool
        :return: True if database was deleted successfully, False if database
            was not found and **ignore_missing** was set to True.
        :rtype: bool
        :raise arango.exceptions.DatabaseDeleteError: If delete fails.
        """
        request = Request(method="delete", endpoint="/_api/database/{}".format(name))

        def response_handler(resp):
            if resp.error_code == 1228 and ignore_missing:
                return False
            if not resp.is_success:
                raise DatabaseDeleteError(resp, request)
            return resp.body["result"]

        return await self._execute(request, response_handler)

    #########################
    # Collection Management #
    #########################

    def collection(self, name):
        """Return the standard collection API wrapper.

        :param name: Collection name.
        :type name: str | unicode
        :return: Standard collection API wrapper.
        :rtype: arango.collection.StandardCollection
        """
        return StandardCollection(self._conn, self._executor, name)

    async def has_collection(self, name):
        """Check if collection exists in the database.

        :param name: Collection name.
        :type name: str | unicode
        :return: True if collection exists, False otherwise.
        :rtype: bool
        """
        return any(col["name"] == name for col in await self.collections())

    async def collections(self):
        """Return the collections in the database.

        :return: Collections in the database and their details.
        :rtype: [dict]
        :raise arango.exceptions.CollectionListError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/collection")

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionListError(resp, request)
            return [
                {
                    "id": col["id"],
                    "name": col["name"],
                    "system": col["isSystem"],
                    "type": StandardCollection.types[col["type"]],
                    "status": StandardCollection.statuses[col["status"]],
                }
                for col in map(dict, resp.body["result"])
            ]

        return await self._execute(request, response_handler)

    async def create_collection(
        self,
        name,
        sync=False,
        compact=True,
        system=False,
        journal_size=None,
        edge=False,
        volatile=False,
        user_keys=True,
        key_increment=None,
        key_offset=None,
        key_generator="traditional",
        shard_fields=None,
        shard_count=None,
        index_bucket_count=None,
        replication_factor=None,
        shard_like=None,
        sync_replication=None,
        enforce_replication_factor=None,
        sharding_strategy=None,
        smart_join_attribute=None,
        write_concern=None,
    ):
        """Create a new collection.

        :param name: Collection name.
        :type name: str | unicode
        :param sync: If set to True, document operations via the collection
            will block until synchronized to disk by default.
        :type sync: bool
        :param compact: If set to True, the collection is compacted. Applies
            only to MMFiles storage engine.
        :type compact: bool
        :param system: If set to True, a system collection is created. The
            collection name must have leading underscore "_" character.
        :type system: bool
        :param journal_size: Max size of the journal in bytes.
        :type journal_size: int
        :param edge: If set to True, an edge collection is created.
        :type edge: bool
        :param volatile: If set to True, collection data is kept in-memory only
            and not made persistent. Unloading the collection will cause the
            collection data to be discarded. Stopping or re-starting the server
            will also cause full loss of data.
        :type volatile: bool
        :param key_generator: Used for generating document keys. Allowed values
            are "traditional" or "autoincrement".
        :type key_generator: str | unicode
        :param user_keys: If set to True, users are allowed to supply document
            keys. If set to False, the key generator is solely responsible for
            supplying the key values.
        :type user_keys: bool
        :param key_increment: Key increment value. Applies only when value of
            **key_generator** is set to "autoincrement".
        :type key_increment: int
        :param key_offset: Key offset value. Applies only when value of
            **key_generator** is set to "autoincrement".
        :type key_offset: int
        :param shard_fields: Field(s) used to determine the target shard.
        :type shard_fields: [str | unicode]
        :param shard_count: Number of shards to create.
        :type shard_count: int
        :param index_bucket_count: Number of buckets into which indexes using
            hash tables are split. The default is 16, and this number has to be
            a power of 2 and less than or equal to 1024. For large collections,
            one should increase this to avoid long pauses when the hash table
            has to be initially built or re-sized, since buckets are re-sized
            individually and can be initially built in parallel. For instance,
            64 may be a sensible value for 100 million documents.
        :type index_bucket_count: int
        :param replication_factor: Number of copies of each shard on different
            servers in a cluster. Allowed values are 1 (only one copy is kept
            and no synchronous replication), and n (n-1 replicas are kept and
            any two copies are replicated across servers synchronously, meaning
            every write to the master is copied to all slaves before operation
            is reported successful).
        :type replication_factor: int
        :param shard_like: Name of prototype collection whose sharding
            specifics are imitated. Prototype collections cannot be dropped
            before imitating collections. Applies to enterprise version of
            ArangoDB only.
        :type shard_like: str | unicode
        :param sync_replication: If set to True, server reports success only
            when collection is created in all replicas. You can set this to
            False for faster server response, and if full replication is not a
            concern.
        :type sync_replication: bool
        :param enforce_replication_factor: Check if there are enough replicas
            available at creation time, or halt the operation.
        :type enforce_replication_factor: bool
        :param sharding_strategy: Sharding strategy. Available for ArangoDB
            version  and up only. Possible values are "community-compat",
            "enterprise-compat", "enterprise-smart-edge-compat", "hash" and
            "enterprise-hash-smart-edge". Refer to ArangoDB documentation for
            more details on each value.
        :type sharding_strategy: str | unicode
        :param smart_join_attribute: Attribute of the collection which must
            contain the shard key value of the smart join collection. The shard
            key for the documents must contain the value of this attribute,
            followed by a colon ":" and the primary key of the document.
            Requires parameter **shard_like** to be set to the name of another
            collection, and parameter **shard_fields** to be set to a single
            shard key attribute, with another colon ":" at the end. Available
            only for enterprise version of ArangoDB.
        :type smart_join_attribute: str | unicode
        :param write_concern: Write concern for the collection. Determines how
            many copies of each shard are required to be in sync on different
            DBServers. If there are less than these many copies in the cluster
            a shard will refuse to write. Writes to shards with enough
            up-to-date copies will succeed at the same time. The value of this
            parameter cannot be larger than that of **replication_factor**.
            Default value is 1. Used for clusters only.
        :type write_concern: int
        :return: Standard collection API wrapper.
        :rtype: arango.collection.StandardCollection
        :raise arango.exceptions.CollectionCreateError: If create fails.
        """
        key_options = {"type": key_generator, "allowUserKeys": user_keys}
        if key_increment is not None:
            key_options["increment"] = key_increment
        if key_offset is not None:
            key_options["offset"] = key_offset

        data = {
            "name": name,
            "waitForSync": sync,
            "doCompact": compact,
            "isSystem": system,
            "isVolatile": volatile,
            "keyOptions": key_options,
            "type": 3 if edge else 2,
        }
        if journal_size is not None:
            data["journalSize"] = journal_size
        if shard_count is not None:
            data["numberOfShards"] = shard_count
        if shard_fields is not None:
            data["shardKeys"] = shard_fields
        if index_bucket_count is not None:
            data["indexBuckets"] = index_bucket_count
        if replication_factor is not None:
            data["replicationFactor"] = replication_factor
        if shard_like is not None:
            data["distributeShardsLike"] = shard_like
        if sharding_strategy is not None:
            data["shardingStrategy"] = sharding_strategy
        if smart_join_attribute is not None:
            data["smartJoinAttribute"] = smart_join_attribute
        if write_concern is not None:
            data["writeConcern"] = write_concern

        params = {}
        if sync_replication is not None:
            params["waitForSyncReplication"] = sync_replication
        if enforce_replication_factor is not None:
            params["enforceReplicationFactor"] = enforce_replication_factor

        request = Request(
            method="post", endpoint="/_api/collection", params=params, data=data
        )

        def response_handler(resp):
            if resp.is_success:
                return self.collection(name)
            raise CollectionCreateError(resp, request)

        return await self._execute(request, response_handler)

    async def delete_collection(self, name, ignore_missing=False, system=None):
        """Delete the collection.

        :param name: Collection name.
        :type name: str | unicode
        :param ignore_missing: Do not raise an exception on missing collection.
        :type ignore_missing: bool
        :param system: Whether the collection is a system collection.
        :type system: bool
        :return: True if collection was deleted successfully, False if
            collection was not found and **ignore_missing** was set to True.
        :rtype: bool
        :raise arango.exceptions.CollectionDeleteError: If delete fails.
        """
        params = {}
        if system is not None:
            params["isSystem"] = system

        request = Request(
            method="delete", endpoint="/_api/collection/{}".format(name), params=params
        )

        def response_handler(resp):
            if resp.error_code == 1203 and ignore_missing:
                return False
            if not resp.is_success:
                raise CollectionDeleteError(resp, request)
            return True

        return await self._execute(request, response_handler)

    ####################
    # Graph Management #
    ####################

    def graph(self, name):
        """Return the graph API wrapper.

        :param name: Graph name.
        :type name: str | unicode
        :return: Graph API wrapper.
        :rtype: arango.graph.Graph
        """
        return Graph(self._conn, self._executor, name)

    async def has_graph(self, name):
        """Check if a graph exists in the database.

        :param name: Graph name.
        :type name: str | unicode
        :return: True if graph exists, False otherwise.
        :rtype: bool
        """
        for graph in await self.graphs():
            if graph["name"] == name:
                return True
        return False

    async def graphs(self):
        """List all graphs in the database.

        :return: Graphs in the database.
        :rtype: [dict]
        :raise arango.exceptions.GraphListError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/gharial")

        def response_handler(resp):
            if not resp.is_success:
                raise GraphListError(resp, request)
            return [
                {
                    "id": body["_id"],
                    "name": body["_key"],
                    "revision": body["_rev"],
                    "orphan_collections": body["orphanCollections"],
                    "edge_definitions": [
                        {
                            "edge_collection": definition["collection"],
                            "from_vertex_collections": definition["from"],
                            "to_vertex_collections": definition["to"],
                        }
                        for definition in body["edgeDefinitions"]
                    ],
                    "shard_count": body.get("numberOfShards"),
                    "replication_factor": body.get("replicationFactor"),
                }
                for body in resp.body["graphs"]
            ]

        return await self._execute(request, response_handler)

    async def create_graph(
        self,
        name,
        edge_definitions=None,
        orphan_collections=None,
        smart=None,
        smart_field=None,
        shard_count=None,
    ):
        """Create a new graph.

        :param name: Graph name.
        :type name: str | unicode
        :param edge_definitions: List of edge definitions, where each edge
            definition entry is a dictionary with fields "edge_collection",
            "from_vertex_collections" and "to_vertex_collections" (see below
            for example).
        :type edge_definitions: [dict]
        :param orphan_collections: Names of additional vertex collections that
            are not in edge definitions.
        :type orphan_collections: [str | unicode]
        :param smart: If set to True, sharding is enabled (see parameter
            **smart_field** below). Applies only to enterprise version of
            ArangoDB.
        :type smart: bool
        :param smart_field: Document field used to shard the vertices of the
            graph. To use this, parameter **smart** must be set to True and
            every vertex in the graph must have the smart field. Applies only
            to enterprise version of ArangoDB.
        :type smart_field: str | unicode
        :param shard_count: Number of shards used for every collection in the
            graph. To use this, parameter **smart** must be set to True and
            every vertex in the graph must have the smart field. This number
            cannot be modified later once set. Applies only to enterprise
            version of ArangoDB.
        :type shard_count: int
        :return: Graph API wrapper.
        :rtype: arango.graph.Graph
        :raise arango.exceptions.GraphCreateError: If create fails.

        Here is an example entry for parameter **edge_definitions**:

        .. code-block:: python

            {
                'edge_collection': 'teach',
                'from_vertex_collections': ['teachers'],
                'to_vertex_collections': ['lectures']
            }
        """
        data = {"name": name}
        if edge_definitions is not None:
            data["edgeDefinitions"] = [
                {
                    "collection": definition["edge_collection"],
                    "from": definition["from_vertex_collections"],
                    "to": definition["to_vertex_collections"],
                }
                for definition in edge_definitions
            ]
        if orphan_collections is not None:
            data["orphanCollections"] = orphan_collections
        if smart is not None:  # pragma: no cover
            data["isSmart"] = smart
        if smart_field is not None:  # pragma: no cover
            data["smartGraphAttribute"] = smart_field
        if shard_count is not None:  # pragma: no cover
            data["numberOfShards"] = shard_count

        request = Request(method="post", endpoint="/_api/gharial", data=data)

        def response_handler(resp):
            if resp.is_success:
                return Graph(self._conn, self._executor, name)
            raise GraphCreateError(resp, request)

        return await self._execute(request, response_handler)

    async def delete_graph(self, name, ignore_missing=False, drop_collections=None):
        """Drop the graph of the given name from the database.

        :param name: Graph name.
        :type name: str | unicode
        :param ignore_missing: Do not raise an exception on missing graph.
        :type ignore_missing: bool
        :param drop_collections: Drop the collections of the graph also. This
            is only if they are not in use by other graphs.
        :type drop_collections: bool
        :return: True if graph was deleted successfully, False if graph was not
            found and **ignore_missing** was set to True.
        :rtype: bool
        :raise arango.exceptions.GraphDeleteError: If delete fails.
        """
        params = {}
        if drop_collections is not None:
            params["dropCollections"] = drop_collections

        request = Request(
            method="delete", endpoint="/_api/gharial/{}".format(name), params=params
        )

        def response_handler(resp):
            if resp.error_code == 1924 and ignore_missing:
                return False
            if not resp.is_success:
                raise GraphDeleteError(resp, request)
            return True

        return await self._execute(request, response_handler)

    #######################
    # Document Management #
    #######################

    async def has_document(self, document, rev=None, check_rev=True):
        """Check if a document exists.

        :param document: Document ID or body with "_id" field.
        :type document: str | unicode | dict
        :param rev: Expected document revision. Overrides value of "_rev" field
            in **document** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **document** (if given)
            is compared against the revision of target document.
        :type check_rev: bool
        :return: True if document exists, False otherwise.
        :rtype: bool
        :raise arango.exceptions.DocumentInError: If check fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_doc(document).has(
            document=document, rev=rev, check_rev=check_rev
        )

    async def document(self, document, rev=None, check_rev=True):
        """Return a document.

        :param document: Document ID or body with "_id" field.
        :type document: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **document** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **document** (if given)
            is compared against the revision of target document.
        :type check_rev: bool
        :return: Document, or None if not found.
        :rtype: dict | None
        :raise arango.exceptions.DocumentGetError: If retrieval fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_doc(document).get(
            document=document, rev=rev, check_rev=check_rev
        )

    async def insert_document(
        self,
        collection,
        document,
        return_new=False,
        sync=None,
        silent=False,
        overwrite=False,
        return_old=False,
        overwrite_mode=None,
        keep_none=None,
        merge=None,
    ):
        """Insert a new document.

        :param collection: Collection name.
        :type collection: str | unicode
        :param document: Document to insert. If it contains the "_key" or "_id"
            field, the value is used as the key of the new document (otherwise
            it is auto-generated). Any "_rev" field is ignored.
        :type document: dict
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :param overwrite: If set to True, operation does not fail on duplicate
            key and the existing document is replaced.
        :type overwrite: bool
        :param return_old: Include body of the old document if replaced.
            Applies only when value of **overwrite** is set to True.
        :type return_old: bool
        :param overwrite_mode: Overwrite behavior used when the document key
            exists already. Allowed values are "replace" (replace-insert) or
            "update" (update-insert). Implicitly sets the value of parameter
            **overwrite**.
        :type overwrite_mode: str
        :param keep_none: If set to True, fields with value None are retained
            in the document. Otherwise, they are removed completely. Applies
            only when **overwrite_mode** is set to "update" (update-insert).
        :type keep_none: bool
        :param merge: If set to True (default), sub-dictionaries are merged
            instead of the new one overwriting the old one. Applies only when
            **overwrite_mode** is set to "update" (update-insert).
        :type merge: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentInsertError: If insert fails.
        """
        return await self.collection(collection).insert(
            document=document,
            return_new=return_new,
            sync=sync,
            silent=silent,
            overwrite=overwrite,
            return_old=return_old,
            overwrite_mode=overwrite_mode,
            keep_none=keep_none,
            merge=merge,
        )

    async def update_document(
        self,
        document,
        check_rev=True,
        merge=True,
        keep_none=True,
        return_new=False,
        return_old=False,
        sync=None,
        silent=False,
    ):
        """Update a document.

        :param document: Partial or full document with the updated values. It
            must contain the "_id" field.
        :type document: dict
        :param check_rev: If set to True, revision of **document** (if given)
            is compared against the revision of target document.
        :type check_rev: bool
        :param merge: If set to True, sub-dictionaries are merged instead of
            the new one overwriting the old one.
        :type merge: bool
        :param keep_none: If set to True, fields with value None are retained
            in the document. Otherwise, they are removed completely.
        :type keep_none: bool
        :param return_new: Include body of the new document in the result.
        :type return_new: bool
        :param return_old: Include body of the old document in the result.
        :type return_old: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentUpdateError: If update fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_doc(document).update(
            document=document,
            check_rev=check_rev,
            merge=merge,
            keep_none=keep_none,
            return_new=return_new,
            return_old=return_old,
            sync=sync,
            silent=silent,
        )

    async def replace_document(
        self,
        document,
        check_rev=True,
        return_new=False,
        return_old=False,
        sync=None,
        silent=False,
    ):
        """Replace a document.

        :param document: New document to replace the old one with. It must
            contain the "_id" field. Edge document must also have "_from" and
            "_to" fields.
        :type document: dict
        :param check_rev: If set to True, revision of **document** (if given)
            is compared against the revision of target document.
        :type check_rev: bool
        :param return_new: Include body of the new document in the result.
        :type return_new: bool
        :param return_old: Include body of the old document in the result.
        :type return_old: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentReplaceError: If replace fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_doc(document).replace(
            document=document,
            check_rev=check_rev,
            return_new=return_new,
            return_old=return_old,
            sync=sync,
            silent=silent,
        )

    async def delete_document(
        self,
        document,
        rev=None,
        check_rev=True,
        ignore_missing=False,
        return_old=False,
        sync=None,
        silent=False,
    ):
        """Delete a document.

        :param document: Document ID, key or body. Document body must contain
            the "_id" field.
        :type document: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **document** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **document** (if given)
            is compared against the revision of target document.
        :type check_rev: bool
        :param ignore_missing: Do not raise an exception on missing document.
            This parameter has no effect in transactions where an exception is
            always raised on failures.
        :type ignore_missing: bool
        :param return_old: Include body of the old document in the result.
        :type return_old: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision), or True if
            parameter **silent** was set to True, or False if document was not
            found and **ignore_missing** was set to True (does not apply in
            transactions).
        :rtype: bool | dict
        :raise arango.exceptions.DocumentDeleteError: If delete fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_doc(document).delete(
            document=document,
            rev=rev,
            check_rev=check_rev,
            ignore_missing=ignore_missing,
            return_old=return_old,
            sync=sync,
            silent=silent,
        )

    ###################
    # Task Management #
    ###################

    async def tasks(self):
        """Return all currently active server tasks.

        :return: Currently active server tasks.
        :rtype: [dict]
        :raise arango.exceptions.TaskListError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/tasks")

        def response_handler(resp):
            if not resp.is_success:
                raise TaskListError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def task(self, task_id):
        """Return the details of an active server task.

        :param task_id: Server task ID.
        :type task_id: str | unicode
        :return: Server task details.
        :rtype: dict
        :raise arango.exceptions.TaskGetError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/tasks/{}".format(task_id))

        def response_handler(resp):
            if not resp.is_success:
                raise TaskGetError(resp, request)
            resp.body.pop("code", None)
            resp.body.pop("error", None)
            return resp.body

        return await self._execute(request, response_handler)

    async def create_task(
        self, name, command, params=None, period=None, offset=None, task_id=None
    ):
        """Create a new server task.

        :param name: Name of the server task.
        :type name: str | unicode
        :param command: Javascript command to execute.
        :type command: str | unicode
        :param params: Optional parameters passed into the Javascript command.
        :type params: dict
        :param period: Number of seconds to wait between executions. If set
            to 0, the new task will be "timed", meaning it will execute only
            once and be deleted afterwards.
        :type period: int
        :param offset: Initial delay before execution in seconds.
        :type offset: int
        :param task_id: Pre-defined ID for the new server task.
        :type task_id: str | unicode
        :return: Details of the new task.
        :rtype: dict
        :raise arango.exceptions.TaskCreateError: If create fails.
        """
        data = {"name": name, "command": command}
        if params is not None:
            data["params"] = params
        if task_id is not None:
            data["id"] = task_id
        if period is not None:
            data["period"] = period
        if offset is not None:
            data["offset"] = offset

        if task_id is None:
            task_id = ""

        request = Request(
            method="post", endpoint="/_api/tasks/{}".format(task_id), data=data
        )

        def response_handler(resp):
            if not resp.is_success:
                raise TaskCreateError(resp, request)
            resp.body.pop("code", None)
            resp.body.pop("error", None)
            return resp.body

        return await self._execute(request, response_handler)

    async def delete_task(self, task_id, ignore_missing=False):
        """Delete a server task.

        :param task_id: Server task ID.
        :type task_id: str | unicode
        :param ignore_missing: Do not raise an exception on missing task.
        :type ignore_missing: bool
        :return: True if task was successfully deleted, False if task was not
            found and **ignore_missing** was set to True.
        :rtype: bool
        :raise arango.exceptions.TaskDeleteError: If delete fails.
        """
        request = Request(method="delete", endpoint="/_api/tasks/{}".format(task_id))

        def response_handler(resp):
            if resp.error_code == 1852 and ignore_missing:
                return False
            if not resp.is_success:
                raise TaskDeleteError(resp, request)
            return True

        return await self._execute(request, response_handler)

    ###################
    # User Management #
    ###################

    async def has_user(self, username):
        """Check if user exists.

        :param username: Username.
        :type username: str | unicode
        :return: True if user exists, False otherwise.
        :rtype: bool
        """
        return any(user["username"] == username for user in await self.users())

    async def users(self):
        """Return all user details.

        :return: List of user details.
        :rtype: [dict]
        :raise arango.exceptions.UserListError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/user")

        def response_handler(resp):
            if not resp.is_success:
                raise UserListError(resp, request)
            return [
                {
                    "username": record["user"],
                    "active": record["active"],
                    "extra": record["extra"],
                }
                for record in resp.body["result"]
            ]

        return await self._execute(request, response_handler)

    async def user(self, username):
        """Return user details.

        :param username: Username.
        :type username: str | unicode
        :return: User details.
        :rtype: dict
        :raise arango.exceptions.UserGetError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/user/{}".format(username))

        def response_handler(resp):
            if not resp.is_success:
                raise UserGetError(resp, request)
            return {
                "username": resp.body["user"],
                "active": resp.body["active"],
                "extra": resp.body["extra"],
            }

        return await self._execute(request, response_handler)

    async def create_user(self, username, password, active=True, extra=None):
        """Create a new user.

        :param username: Username.
        :type username: str | unicode
        :param password: Password.
        :type password: str | unicode
        :param active: True if user is active, False otherwise.
        :type active: bool
        :param extra: Additional data for the user.
        :type extra: dict
        :return: New user details.
        :rtype: dict
        :raise arango.exceptions.UserCreateError: If create fails.
        """
        data = {"user": username, "passwd": password, "active": active}
        if extra is not None:
            data["extra"] = extra

        request = Request(method="post", endpoint="/_api/user", data=data)

        def response_handler(resp):
            if not resp.is_success:
                raise UserCreateError(resp, request)
            return {
                "username": resp.body["user"],
                "active": resp.body["active"],
                "extra": resp.body["extra"],
            }

        return await self._execute(request, response_handler)

    async def update_user(self, username, password=None, active=None, extra=None):
        """Update a user.

        :param username: Username.
        :type username: str | unicode
        :param password: New password.
        :type password: str | unicode
        :param active: Whether the user is active.
        :type active: bool
        :param extra: Additional data for the user.
        :type extra: dict
        :return: New user details.
        :rtype: dict
        :raise arango.exceptions.UserUpdateError: If update fails.
        """
        data = {}
        if password is not None:
            data["passwd"] = password
        if active is not None:
            data["active"] = active
        if extra is not None:
            data["extra"] = extra

        request = Request(
            method="patch",
            endpoint="/_api/user/{user}".format(user=username),
            data=data,
        )

        def response_handler(resp):
            if not resp.is_success:
                raise UserUpdateError(resp, request)
            return {
                "username": resp.body["user"],
                "active": resp.body["active"],
                "extra": resp.body["extra"],
            }

        return await self._execute(request, response_handler)

    async def replace_user(self, username, password, active=None, extra=None):
        """Replace a user.

        :param username: Username.
        :type username: str | unicode
        :param password: New password.
        :type password: str | unicode
        :param active: Whether the user is active.
        :type active: bool
        :param extra: Additional data for the user.
        :type extra: dict
        :return: New user details.
        :rtype: dict
        :raise arango.exceptions.UserReplaceError: If replace fails.
        """
        data = {"user": username, "passwd": password}
        if active is not None:
            data["active"] = active
        if extra is not None:
            data["extra"] = extra

        request = Request(
            method="put", endpoint="/_api/user/{user}".format(user=username), data=data
        )

        def response_handler(resp):
            if resp.is_success:
                return {
                    "username": resp.body["user"],
                    "active": resp.body["active"],
                    "extra": resp.body["extra"],
                }
            raise UserReplaceError(resp, request)

        return await self._execute(request, response_handler)

    async def delete_user(self, username, ignore_missing=False):
        """Delete a user.

        :param username: Username.
        :type username: str | unicode
        :param ignore_missing: Do not raise an exception on missing user.
        :type ignore_missing: bool
        :return: True if user was deleted successfully, False if user was not
            found and **ignore_missing** was set to True.
        :rtype: bool
        :raise arango.exceptions.UserDeleteError: If delete fails.
        """
        request = Request(
            method="delete", endpoint="/_api/user/{user}".format(user=username)
        )

        def response_handler(resp):
            if resp.is_success:
                return True
            elif resp.status_code == 404 and ignore_missing:
                return False
            raise UserDeleteError(resp, request)

        return await self._execute(request, response_handler)

    #########################
    # Permission Management #
    #########################

    async def permissions(self, username):
        """Return user permissions for all databases and collections.

        :param username: Username.
        :type username: str | unicode
        :return: User permissions for all databases and collections.
        :rtype: dict
        :raise arango.exceptions.PermissionListError: If retrieval fails.
        """
        request = Request(
            method="get",
            endpoint="/_api/user/{}/database".format(username),
            params={"full": True},
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body["result"]
            raise PermissionListError(resp, request)

        return await self._execute(request, response_handler)

    async def permission(self, username, database, collection=None):
        """Return user permission for a specific database or collection.

        :param username: Username.
        :type username: str | unicode
        :param database: Database name.
        :type database: str | unicode
        :param collection: Collection name.
        :type collection: str | unicode
        :return: Permission for given database or collection.
        :rtype: str | unicode
        :raise arango.exceptions.PermissionGetError: If retrieval fails.
        """
        endpoint = "/_api/user/{}/database/{}".format(username, database)
        if collection is not None:
            endpoint += "/" + collection
        request = Request(method="get", endpoint=endpoint)

        def response_handler(resp):
            if not resp.is_success:
                raise PermissionGetError(resp, request)
            return resp.body["result"]

        return await self._execute(request, response_handler)

    async def update_permission(self, username, permission, database, collection=None):
        """Update user permission for a specific database or collection.

        :param username: Username.
        :type username: str | unicode
        :param database: Database name.
        :type database: str | unicode
        :param collection: Collection name.
        :type collection: str | unicode
        :param permission: Allowed values are "rw" (read and write), "ro"
            (read only) or "none" (no access).
        :type permission: str | unicode
        :return: True if access was granted successfully.
        :rtype: bool
        :raise arango.exceptions.PermissionUpdateError: If update fails.
        """
        endpoint = "/_api/user/{}/database/{}".format(username, database)
        if collection is not None:
            endpoint += "/" + collection

        request = Request(method="put", endpoint=endpoint, data={"grant": permission})

        def response_handler(resp):
            if resp.is_success:
                return True
            raise PermissionUpdateError(resp, request)

        return await self._execute(request, response_handler)

    async def reset_permission(self, username, database, collection=None):
        """Reset user permission for a specific database or collection.

        :param username: Username.
        :type username: str | unicode
        :param database: Database name.
        :type database: str | unicode
        :param collection: Collection name.
        :type collection: str | unicode
        :return: True if permission was reset successfully.
        :rtype: bool
        :raise arango.exceptions.PermissionRestError: If reset fails.
        """
        endpoint = "/_api/user/{}/database/{}".format(username, database)
        if collection is not None:
            endpoint += "/" + collection

        request = Request(method="delete", endpoint=endpoint)

        def response_handler(resp):
            if resp.is_success:
                return True
            raise PermissionResetError(resp, request)

        return await self._execute(request, response_handler)

    ########################
    # Async Job Management #
    ########################

    async def async_jobs(self, status, count=None):
        """Return IDs of async jobs with given status.

        :param status: Job status (e.g. "pending", "done").
        :type status: str | unicode
        :param count: Max number of job IDs to return.
        :type count: int
        :return: List of job IDs.
        :rtype: [str | unicode]
        :raise arango.exceptions.AsyncJobListError: If retrieval fails.
        """
        params = {}
        if count is not None:
            params["count"] = count

        request = Request(
            method="get", endpoint="/_api/job/{}".format(status), params=params
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body
            raise AsyncJobListError(resp, request)

        return await self._execute(request, response_handler)

    async def clear_async_jobs(self, threshold=None):
        """Clear async job results from the server.

        Async jobs that are still queued or running are not stopped.

        :param threshold: If specified, only the job results created prior to
            the threshold (a unix timestamp) are deleted. Otherwise, all job
            results are deleted.
        :type threshold: int
        :return: True if job results were cleared successfully.
        :rtype: bool
        :raise arango.exceptions.AsyncJobClearError: If operation fails.
        """
        if threshold is None:
            url = "/_api/job/all"
            params = None
        else:
            url = "/_api/job/expired"
            params = {"stamp": threshold}

        request = Request(method="delete", endpoint=url, params=params)

        def response_handler(resp):
            if resp.is_success:
                return True
            raise AsyncJobClearError(resp, request)

        return await self._execute(request, response_handler)

    ###################
    # View Management #
    ###################

    async def views(self):
        """Return list of views and their summaries.

        :return: List of views.
        :rtype: [dict]
        :raise arango.exceptions.ViewListError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/view")

        def response_handler(resp):
            if resp.is_success:
                return [format_view(view) for view in resp.body["result"]]
            raise ViewListError(resp, request)

        return await self._execute(request, response_handler)

    async def view(self, name):
        """Return view details.

        :return: View details.
        :rtype: dict
        :raise arango.exceptions.ViewGetError: If retrieval fails.
        """
        request = Request(
            method="get", endpoint="/_api/view/{}/properties".format(name)
        )

        def response_handler(resp):
            if resp.is_success:
                return format_view(resp.body)
            raise ViewGetError(resp, request)

        return await self._execute(request, response_handler)

    async def create_view(self, name, view_type, properties=None):
        """Create a view.

        :param name: View name.
        :type name: str | unicode
        :param view_type: View type (e.g. "arangosearch").
        :type view_type: str | unicode
        :param properties: View properties. For more information see
            https://www.arangodb.com/docs/stable/http/views-arangosearch.html
        :type properties: dict
        :return: View details.
        :rtype: dict
        :raise arango.exceptions.ViewCreateError: If create fails.
        """
        data = {"name": name, "type": view_type}

        if properties is not None:
            data.update(properties)

        request = Request(method="post", endpoint="/_api/view", data=data)

        def response_handler(resp):
            if resp.is_success:
                return format_view(resp.body)
            raise ViewCreateError(resp, request)

        return await self._execute(request, response_handler)

    async def update_view(self, name, properties):
        """Update a view.

        :param name: View name.
        :type name: str | unicode
        :param properties: View properties. For more information see
            https://www.arangodb.com/docs/stable/http/views-arangosearch.html
        :type properties: dict
        :return: View details.
        :rtype: dict
        :raise arango.exceptions.ViewUpdateError: If update fails.
        """
        request = Request(
            method="patch",
            endpoint="/_api/view/{}/properties".format(name),
            data=properties,
        )

        def response_handler(resp):
            if resp.is_success:
                return format_view(resp.body)
            raise ViewUpdateError(resp, request)

        return await self._execute(request, response_handler)

    async def replace_view(self, name, properties):
        """Replace a view.

        :param name: View name.
        :type name: str | unicode
        :param properties: View properties. For more information see
            https://www.arangodb.com/docs/stable/http/views-arangosearch.html
        :type properties: dict
        :return: View details.
        :rtype: dict
        :raise arango.exceptions.ViewReplaceError: If replace fails.
        """
        request = Request(
            method="put",
            endpoint="/_api/view/{}/properties".format(name),
            data=properties,
        )

        def response_handler(resp):
            if resp.is_success:
                return format_view(resp.body)
            raise ViewReplaceError(resp, request)

        return await self._execute(request, response_handler)

    async def delete_view(self, name, ignore_missing=False):
        """Delete a view.

        :param name: View name.
        :type name: str | unicode
        :param ignore_missing: Do not raise an exception on missing view.
        :type ignore_missing: bool
        :return: True if view was deleted successfully, False if view was not
            found and **ignore_missing** was set to True.
        :rtype: bool
        :raise arango.exceptions.ViewDeleteError: If delete fails.
        """
        request = Request(method="delete", endpoint="/_api/view/{}".format(name))

        def response_handler(resp):
            if resp.error_code == 1203 and ignore_missing:
                return False
            if resp.is_success:
                return True
            raise ViewDeleteError(resp, request)

        return await self._execute(request, response_handler)

    async def rename_view(self, name, new_name):
        """Rename a view.

        :param name: View name.
        :type name: str | unicode
        :param new_name: New view name.
        :type new_name: str | unicode
        :return: View details.
        :rtype: dict
        :raise arango.exceptions.ViewRenameError: If delete fails.
        """
        request = Request(
            method="put",
            endpoint="/_api/view/{}/rename".format(name),
            data={"name": new_name},
        )

        def response_handler(resp):
            if resp.is_success:
                return True
            raise ViewRenameError(resp, request)

        return await self._execute(request, response_handler)

    ################################
    # ArangoSearch View Management #
    ################################

    async def create_arangosearch_view(self, name, properties=None):
        """Create an ArangoSearch view.

        :param name: View name.
        :type name: str | unicode
        :param properties: View properties. For more information see
            https://www.arangodb.com/docs/stable/http/views-arangosearch.html
        :type properties: dict
        :return: View details.
        :rtype: dict
        :raise arango.exceptions.ViewCreateError: If create fails.
        """
        data = {"name": name, "type": "arangosearch"}

        if properties is not None:
            data.update(properties)

        request = Request(method="post", endpoint="/_api/view#ArangoSearch", data=data)

        def response_handler(resp):
            if resp.is_success:
                return format_view(resp.body)
            raise ViewCreateError(resp, request)

        return await self._execute(request, response_handler)

    async def update_arangosearch_view(self, name, properties):
        """Update an ArangoSearch view.

        :param name: View name.
        :type name: str | unicode
        :param properties: View properties. For more information see
            https://www.arangodb.com/docs/stable/http/views-arangosearch.html
        :type properties: dict
        :return: View details.
        :rtype: dict
        :raise arango.exceptions.ViewUpdateError: If update fails.
        """
        request = Request(
            method="patch",
            endpoint="/_api/view/{}/properties#ArangoSearch".format(name),
            data=properties,
        )

        def response_handler(resp):
            if resp.is_success:
                return format_view(resp.body)
            raise ViewUpdateError(resp, request)

        return await self._execute(request, response_handler)

    async def replace_arangosearch_view(self, name, properties):
        """Replace an ArangoSearch view.

        :param name: View name.
        :type name: str | unicode
        :param properties: View properties. For more information see
            https://www.arangodb.com/docs/stable/http/views-arangosearch.html
        :type properties: dict
        :return: View details.
        :rtype: dict
        :raise arango.exceptions.ViewReplaceError: If replace fails.
        """
        request = Request(
            method="put",
            endpoint="/_api/view/{}/properties#ArangoSearch".format(name),
            data=properties,
        )

        def response_handler(resp):
            if resp.is_success:
                return format_view(resp.body)
            raise ViewReplaceError(resp, request)

        return await self._execute(request, response_handler)

    #######################
    # Analyzer Management #
    #######################

    async def analyzers(self):
        """Return list of analyzers.

        :return: List of analyzers.
        :rtype: [dict]
        :raise arango.exceptions.AnalyzerListError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/analyzer")

        def response_handler(resp):
            if resp.is_success:
                resp.body.pop("error")
                resp.body.pop("code")
                return resp.body["result"]
            raise AnalyzerListError(resp, request)

        return await self._execute(request, response_handler)

    async def analyzer(self, name):
        """Return analyzer details.

        :param name: Analyzer name.
        :type name: str | unicode
        :return: Analyzer details.
        :rtype: dict
        :raise arango.exceptions.AnalyzerGetError: If retrieval fails.
        """
        request = Request(method="get", endpoint="/_api/analyzer/{}".format(name))

        def response_handler(resp):
            if resp.is_success:
                resp.body.pop("error")
                resp.body.pop("code")
                return resp.body
            raise AnalyzerGetError(resp, request)

        return await self._execute(request, response_handler)

    async def create_analyzer(
        self, name, analyzer_type, properties=None, features=None
    ):
        """Create an analyzer.

        :param name: Analyzer name.
        :type name: str | unicode
        :param analyzer_type: Analyzer type.
        :type analyzer_type: str | unicode
        :param properties: Analyzer properties.
        :type properties: dict
        :param features: Analyzer features.
        :type features: list
        :return: Analyzer details.
        :rtype: dict
        :raise arango.exceptions.AnalyzerCreateError: If create fails.
        """
        data = {"name": name, "type": analyzer_type}

        if properties is not None:
            data["properties"] = properties

        if features is not None:
            data["features"] = features

        request = Request(method="post", endpoint="/_api/analyzer", data=data)

        def response_handler(resp):
            if resp.is_success:
                return resp.body
            raise AnalyzerCreateError(resp, request)

        return await self._execute(request, response_handler)

    async def delete_analyzer(self, name, force=False, ignore_missing=False):
        """Delete an analyzer.

        :param name: Analyzer name.
        :type name: str | unicode
        :param force: Remove the analyzer configuration even if in use.
        :type force: bool
        :param ignore_missing: Do not raise an exception on missing analyzer.
        :type ignore_missing: bool
        :return: True if analyzer was deleted successfully, False if analyzer
            was not found and **ignore_missing** was set to True.
        :rtype: bool
        :raise arango.exceptions.AnalyzerDeleteError: If delete fails.
        """
        request = Request(
            method="delete",
            endpoint="/_api/analyzer/{}".format(name),
            params={"force": force},
        )

        def response_handler(resp):
            if resp.error_code in {1202, 404} and ignore_missing:
                return False
            if resp.is_success:
                return True
            raise AnalyzerDeleteError(resp, request)

        return await self._execute(request, response_handler)


class StandardDatabase(Database):
    """Standard database API wrapper.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    """

    def __init__(self, connection):
        super(StandardDatabase, self).__init__(
            connection=connection, executor=DefaultExecutor(connection)
        )

    def __repr__(self):
        return "<StandardDatabase {}>".format(self.name)

    def begin_async_execution(self, return_result=True):
        """Begin async execution.

        :param return_result: If set to True, API executions return instances
            of :class:`arango.job.AsyncJob`, which you can use to retrieve
            results from server once available. If set to False, API executions
            return None and no results are stored on server.
        :type return_result: bool
        :return: Database API wrapper object specifically for async execution.
        :rtype: arango.database.AsyncDatabase
        """
        return AsyncDatabase(self._conn, return_result)

    def begin_batch_execution(self, return_result=True):
        """Begin batch execution.

        :param return_result: If set to True, API executions return instances
            of :class:`arango.job.BatchJob` that are populated with results on
            commit. If set to False, API executions return None and no results
            are tracked client-side.
        :type return_result: bool
        :return: Database API wrapper object specifically for batch execution.
        :rtype: arango.database.BatchDatabase
        """
        return BatchDatabase(self._conn, return_result)

    async def begin_transaction(
        self,
        read=None,
        write=None,
        exclusive=None,
        sync=None,
        allow_implicit=None,
        lock_timeout=None,
        max_size=None,
    ):
        """Begin a transaction.

        :param read: Name(s) of collections read during transaction. Read-only
            collections are added lazily but should be declared if possible to
            avoid deadlocks.
        :type read: str | unicode | [str | unicode]
        :param write: Name(s) of collections written to during transaction with
            shared access.
        :type write: str | unicode | [str | unicode]
        :param exclusive: Name(s) of collections written to during transaction
            with exclusive access.
        :type exclusive: str | unicode | [str | unicode]
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param allow_implicit: Allow reading from undeclared collections.
        :type allow_implicit: bool
        :param lock_timeout: Timeout for waiting on collection locks. If not
            given, a default value is used. Setting it to 0 disables the
            timeout.
        :type lock_timeout: int
        :param max_size: Max transaction size in bytes. Applicable to RocksDB
            storage engine only.
        :type max_size:
        :return: Database API wrapper object specifically for transactions.
        :rtype: arango.database.TransactionDatabase
        """
        tb = TransactionDatabase(connection=self._conn,)
        await tb.begin_transaction(
            read=read,
            write=write,
            exclusive=exclusive,
            sync=sync,
            allow_implicit=allow_implicit,
            lock_timeout=lock_timeout,
            max_size=max_size,
        )
        return tb


class AsyncDatabase(Database):
    """Database API wrapper tailored specifically for async execution.

    See :func:`arango.database.StandardDatabase.begin_async_execution`.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param return_result: If set to True, API executions return instances of
        :class:`arango.job.AsyncJob`, which you can use to retrieve results
        from server once available. If set to False, API executions return None
        and no results are stored on server.
    :type return_result: bool
    """

    def __init__(self, connection, return_result):
        super(AsyncDatabase, self).__init__(
            connection=connection, executor=AsyncExecutor(connection, return_result)
        )

    def __repr__(self):
        return "<AsyncDatabase {}>".format(self.name)


class BatchDatabase(Database):
    """Database API wrapper tailored specifically for batch execution.

    See :func:`arango.database.StandardDatabase.begin_batch_execution`.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param return_result: If set to True, API executions return instances of
        :class:`arango.job.BatchJob` that are populated with results on commit.
        If set to False, API executions return None and no results are tracked
        client-side.
    :type return_result: bool
    """

    def __init__(self, connection, return_result):
        super(BatchDatabase, self).__init__(
            connection=connection, executor=BatchExecutor(connection, return_result)
        )

    def __repr__(self):
        return "<BatchDatabase {}>".format(self.name)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exception, *_):
        if exception is None:
            await self._executor.commit()

    def queued_jobs(self):
        """Return the queued batch jobs.

        :return: Queued batch jobs or None if **return_result** parameter was
            set to False during initialization.
        :rtype: [arango.job.BatchJob] | None
        """
        return self._executor.jobs

    async def commit(self):
        """Execute the queued requests in a single batch API request.

        If **return_result** parameter was set to True during initialization,
        :class:`arango.job.BatchJob` instances are populated with results.

        :return: Batch jobs, or None if **return_result** parameter was set to
            False during initialization.
        :rtype: [arango.job.BatchJob] | None
        :raise arango.exceptions.BatchStateError: If batch state is invalid
            (e.g. batch was already committed or the response size did not
            match expected).
        :raise arango.exceptions.BatchExecuteError: If commit fails.
        """
        return await self._executor.commit()


class TransactionDatabase(Database):
    """Database API wrapper tailored specifically for transactions.

    See :func:`arango.database.StandardDatabase.begin_transaction`.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param read: Name(s) of collections read during transaction. Read-only
        collections are added lazily but should be declared if possible to
        avoid deadlocks.
    :type read: str | unicode | [str | unicode]
    :param write: Name(s) of collections written to during transaction with
        shared access.
    :type write: str | unicode | [str | unicode]
    :param exclusive: Name(s) of collections written to during transaction
        with exclusive access.
    :type exclusive: str | unicode | [str | unicode]
    :param sync: Block until operation is synchronized to disk.
    :type sync: bool
    :param allow_implicit: Allow reading from undeclared collections.
    :type allow_implicit: bool
    :param lock_timeout: Timeout for waiting on collection locks. If not given,
        a default value is used. Setting it to 0 disables the timeout.
    :type lock_timeout: int
    :param max_size: Max transaction size in bytes. Applicable to RocksDB
        storage engine only.
    :type max_size: int
    """

    def __init__(
        self,
        connection,
        read=None,
        write=None,
        exclusive=None,
        sync=None,
        allow_implicit=None,
        lock_timeout=None,
        max_size=None,
    ):
        super(TransactionDatabase, self).__init__(
            connection=connection,
            executor=TransactionExecutor(
                connection=connection
            ),
        )

    def __repr__(self):
        return "<TransactionDatabase {}>".format(self.name)

    @property
    def transaction_id(self):
        """Return the transaction ID.

        :return: Transaction ID.
        :rtype: str | unicode
        """
        return self._executor.id

    async def transaction_status(self):
        """Return the transaction status.

        :return: Transaction status.
        :rtype: str | unicode
        :raise arango.exceptions.TransactionStatusError: If retrieval fails.
        """
        return await self._executor.status()

    async def begin_transaction(
            self,
            read=None,
            write=None,
            exclusive=None,
            sync=None,
            allow_implicit=None,
            lock_timeout=None,
            max_size=None):
        """Commit the transaction.

        :return: True if commit was successful.
        :rtype: bool
        :raise arango.exceptions.TransactionCommitError: If commit fails.
        """
        return await self._executor.begin(
            read=read,
            write=write,
            exclusive=exclusive,
            sync=sync,
            allow_implicit=allow_implicit,
            lock_timeout=lock_timeout,
            max_size=max_size)

    async def commit_transaction(self):
        """Commit the transaction.

        :return: True if commit was successful.
        :rtype: bool
        :raise arango.exceptions.TransactionCommitError: If commit fails.
        """
        return await self._executor.commit()

    async def abort_transaction(self):
        """Abort the transaction.

        :return: True if the abort operation was successful.
        :rtype: bool
        :raise arango.exceptions.TransactionAbortError: If abort fails.
        """
        return await self._executor.abort()



from __future__ import absolute_import, unicode_literals

from abc import ABCMeta, abstractmethod
from calendar import timegm
from datetime import datetime

import jwt
from six import string_types
from aiohttp import MultipartWriter

from .exceptions import (
    ServerConnectionError,
    JWTAuthError,
)
from .request import Request
from .response import Response

__all__ = [
    'Connection',
    'BasicConnection',
    'JWTConnection',
    'JWTSuperuserConnection'
]


class Connection(object):
    """Base connection to specific ArangoDB database.

    :param hosts: Host URL or list of URLs (coordinators in a cluster).
    :type hosts: [str | unicode]
    :param host_resolver: Host resolver (used for clusters).
    :type host_resolver: arango.resolver.HostResolver
    :param sessions: HTTP session objects per host.
    :type sessions: [requests.Session]
    :param db_name: Database name.
    :type db_name: str | unicode
    :param http_client: User-defined HTTP client.
    :type http_client: arango.http.HTTPClient
    """

    __metaclass__ = ABCMeta

    auth_required_header = 'WWW-Authenticate'

    def __init__(self,
                 hosts,
                 host_resolver,
                 sessions,
                 db_name,
                 http_client,
                 serializer,
                 deserializer):
        self._url_prefixes = ['{}/_db/{}'.format(h, db_name) for h in hosts]
        self._host_resolver = host_resolver
        self._sessions = sessions
        self._db_name = db_name
        self._http = http_client
        self._serializer = serializer
        self._deserializer = deserializer

    @property
    def db_name(self):
        """Return the database name.

        :returns: Database name.
        :rtype: str | unicode
        """
        return self._db_name

    def serialize(self, obj):
        """Serialize the object and return the string.

        :param obj: Object to serialize.
        :type obj: str | unicode | bool | int | list | dict
        :return: Serialized string.
        :rtype: str | unicode
        """
        return self._serializer(obj)

    def deserialize(self, string):
        """De-serialize the string and return the object.

        :param string: String to de-serialize.
        :type string: str | unicode
        :return: De-serialized object.
        :rtype: str | unicode | bool | int | list | dict
        """
        try:
            return self._deserializer(string)
        except (ValueError, TypeError):
            return string

    def prep_response(self, resp, deserialize=True):
        """Populate the response with details and return it.

        :param deserialize: Deserialize the response body.
        :type deserialize: bool
        :param resp: HTTP response.
        :type resp: arango.response.Response
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        if deserialize:
            resp.body = self.deserialize(resp.raw_body)
            if isinstance(resp.body, dict):
                resp.error_code = resp.body.get('errorNum')
                resp.error_message = resp.body.get('errorMessage')
        else:
            resp.body = resp.raw_body

        http_ok = 200 <= resp.status_code < 300
        resp.is_success = http_ok and resp.error_code is None
        return resp

    def prep_bulk_err_response(self, parent_response, body):
        """Build and return a bulk error response.

        :param parent_response: Parent response.
        :type parent_response: arango.response.Response
        :param body: Error response body.
        :type body: dict
        :return: Child bulk error response.
        :rtype: arango.response.Response
        """
        resp = Response(
            method=parent_response.method,
            url=parent_response.url,
            headers=parent_response.headers,
            status_code=parent_response.status_code,
            status_text=parent_response.status_text,
            raw_body=self.serialize(body),
        )
        resp.body = body
        resp.error_code = body['errorNum']
        resp.error_message = body['errorMessage']
        resp.is_success = False
        return resp

    def get_normalized_data(self, request):
        """Normalize request data.

        :param request: HTTP request.
        :type request: arango.request.Request
        :return: Normalized data.
        :rtype: str | unicode | bool | int | list | dict
        """
        if request.data is None:
            return request.data
        elif isinstance(request.data, (string_types, MultipartWriter)):
            return request.data
        else:
            return self.serialize(request.data)

    async def ping(self):
        """Ping the next host to check if connection is established.

        :return: Response status code.
        :rtype: int
        """
        request = Request(
            method='get',
            endpoint='/_api/collection'
        )
        resp = await self.send_request(request)
        if resp.status_code in {401, 403}:
            raise ServerConnectionError('bad username and/or password')
        if not resp.is_success:  # pragma: no cover
            raise ServerConnectionError(
                resp.error_message or 'bad server response')
        return resp.status_code

    @abstractmethod
    async def send_request(self, request):  # pragma: no cover
        """Send an HTTP request to ArangoDB server.

        :param request: HTTP request.
        :type request: arango.request.Request
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        raise NotImplementedError


class BasicConnection(Connection):
    """Connection to specific ArangoDB database using basic authentication.

    :param hosts: Host URL or list of URLs (coordinators in a cluster).
    :type hosts: [str | unicode]
    :param host_resolver: Host resolver (used for clusters).
    :type host_resolver: arango.resolver.HostResolver
    :param sessions: HTTP session objects per host.
    :type sessions: [requests.Session]
    :param db_name: Database name.
    :type db_name: str | unicode
    :param username: Username.
    :type username: str | unicode
    :param password: Password.
    :type password: str | unicode
    :param http_client: User-defined HTTP client.
    :type http_client: arango.http.HTTPClient
    """

    def __init__(self,
                 hosts,
                 host_resolver,
                 sessions,
                 db_name,
                 username,
                 password,
                 http_client,
                 serializer,
                 deserializer):
        super(BasicConnection, self).__init__(
            hosts,
            host_resolver,
            sessions,
            db_name,
            http_client,
            serializer,
            deserializer
        )
        self._username = username
        self._auth = (username, password)

    @property
    def username(self):
        """Return the username.

        :returns: Username.
        :rtype: str | unicode
        """
        return self._username

    async def send_request(self, request):
        """Send an HTTP request to ArangoDB server.

        :param request: HTTP request.
        :type request: arango.request.Request
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        host_index = self._host_resolver.get_host_index()
        resp = await self._http.send_request(
            session=self._sessions[host_index],
            method=request.method,
            url=self._url_prefixes[host_index] + request.endpoint,
            params=request.params,
            data=self.get_normalized_data(request),
            headers=request.headers,
            auth=self._auth
        )
        return self.prep_response(resp, request.deserialize)


class JWTConnection(Connection):
    """Connection to specific ArangoDB database using JWT authentication.

    :param hosts: Host URL or list of URLs (coordinators in a cluster).
    :type hosts: [str | unicode]
    :param host_resolver: Host resolver (used for clusters).
    :type host_resolver: arango.resolver.HostResolver
    :param sessions: HTTP session objects per host.
    :type sessions: [requests.Session]
    :param db_name: Database name.
    :type db_name: str | unicode
    :param username: Username.
    :type username: str | unicode
    :param password: Password.
    :type password: str | unicode
    :param http_client: User-defined HTTP client.
    :type http_client: arango.http.HTTPClient
    """

    def __init__(self,
                 hosts,
                 host_resolver,
                 sessions,
                 db_name,
                 username,
                 password,
                 http_client,
                 serializer,
                 deserializer):
        super(JWTConnection, self).__init__(
            hosts,
            host_resolver,
            sessions,
            db_name,
            http_client,
            serializer,
            deserializer
        )
        self._username = username
        self._password = password
        self.exp_leeway = 0

        self._auth_header = None
        self._token = None
        self._token_exp = None

    async def send_request(self, request):
        """Send an HTTP request to ArangoDB server.

        :param request: HTTP request.
        :type request: arango.request.Request
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        host_index = self._host_resolver.get_host_index()
        request.headers['Authorization'] = self._auth_header

        resp = await self._http.send_request(
            session=self._sessions[host_index],
            method=request.method,
            url=self._url_prefixes[host_index] + request.endpoint,
            params=request.params,
            data=self.get_normalized_data(request),
            headers=request.headers,
        )
        resp = self.prep_response(resp, request.deserialize)

        # Refresh the token and retry on HTTP 401 and error code 11.
        if resp.error_code != 11 or resp.status_code != 401:
            return resp

        now = timegm(datetime.utcnow().utctimetuple())
        if self._token_exp < now - self.exp_leeway:  # pragma: no cover
            return resp

        await self.refresh_token()
        request.headers['Authorization'] = self._auth_header

        resp = await self._http.send_request(
            session=self._sessions[host_index],
            method=request.method,
            url=self._url_prefixes[host_index] + request.endpoint,
            params=request.params,
            data=self.get_normalized_data(request),
            headers=request.headers,
        )
        return self.prep_response(resp, request.deserialize)

    async def refresh_token(self):
        """Get a new JWT token for the current user (cannot be a superuser).

        :return: JWT token.
        :rtype: str | unicode
        """
        request = Request(
            method='post',
            endpoint='/_open/auth',
            data={
                'username': self._username,
                'password': self._password
            }
        )
        host_index = self._host_resolver.get_host_index()
        resp = await self._http.send_request(
            session=self._sessions[host_index],
            method=request.method,
            url=self._url_prefixes[host_index] + request.endpoint,
            data=self.get_normalized_data(request)
        )
        resp = self.prep_response(resp)
        if not resp.is_success:
            raise JWTAuthError(resp, request)

        self._token = resp.body['jwt']
        jwt_payload = jwt.decode(
            self._token,
            issuer='arangodb',
            algorithms=['HS256'],
            options={
                'require_exp': True,
                'require_iat': True,
                'verify_iat': True,
                'verify_exp': True,
                'verify_signature': False
            },
        )
        self._token_exp = jwt_payload['exp']
        self._auth_header = 'bearer {}'.format(self._token)


class JWTSuperuserConnection(Connection):
    """Connection to specific ArangoDB database using superuser JWT.

    :param hosts: Host URL or list of URLs (coordinators in a cluster).
    :type hosts: [str | unicode]
    :param host_resolver: Host resolver (used for clusters).
    :type host_resolver: arango.resolver.HostResolver
    :param sessions: HTTP session objects per host.
    :type sessions: [requests.Session]
    :param db_name: Database name.
    :type db_name: str | unicode
    :param http_client: User-defined HTTP client.
    :type http_client: arango.http.HTTPClient
    :param superuser_token: User generated token for superuser access.
    :type superuser_token: str | unicode
    """

    def __init__(self,
                 hosts,
                 host_resolver,
                 sessions,
                 db_name,
                 http_client,
                 serializer,
                 deserializer,
                 superuser_token):
        super(JWTSuperuserConnection, self).__init__(
            hosts,
            host_resolver,
            sessions,
            db_name,
            http_client,
            serializer,
            deserializer
        )
        self._auth_header = 'bearer {}'.format(superuser_token)

    async def send_request(self, request):
        """Send an HTTP request to ArangoDB server.

        :param request: HTTP request.
        :type request: arango.request.Request
        :return: HTTP response.
        :rtype: arango.response.Response
        """
        host_index = self._host_resolver.get_host_index()
        request.headers['Authorization'] = self._auth_header

        resp = await self._http.send_request(
            session=self._sessions[host_index],
            method=request.method,
            url=self._url_prefixes[host_index] + request.endpoint,
            params=request.params,
            data=self.get_normalized_data(request),
            headers=request.headers,
        )
        return self.prep_response(resp, request.deserialize)



from pytest_docker_fixtures.containers._base import BaseImage
from pytest_docker_fixtures.images import settings
import requests


settings['arango'] = {
    'image': 'arangodb',
    'version': '3.6.4',
    'env': {
        'ARANGO_ROOT_PASSWORD': 'secret',
        'ARANGO_NO_AUTH': '1'
    }
}

class ArangoDB(BaseImage):
    name = 'arango'
    port = 8529

    def check(self):
        url = f'http://{self.host}:{self.get_port()}/'
        try:
            resp = requests.get(url)
            if resp.status_code == 200:
                return True
        except Exception:
            pass
        return False


arango_image = ArangoDB()


from __future__ import absolute_import, unicode_literals

import pytest
from aioarangodb.response import Response
pytestmark = pytest.mark.asyncio


async def test_response(conn):
    response = Response(
        method='get',
        url='test_url',
        headers={'foo': 'bar'},
        status_text='baz',
        status_code=200,
        raw_body='true',
    )
    conn.prep_response(response)

    assert response.method == 'get'
    assert response.url == 'test_url'
    assert response.headers == {'foo': 'bar'}
    assert response.status_code == 200
    assert response.status_text == 'baz'
    assert response.raw_body == 'true'
    assert response.body is True
    assert response.error_code is None
    assert response.error_message is None
    assert response.is_success is True

    test_body = '{"errorNum": 1, "errorMessage": "qux"}'
    response = Response(
        method='get',
        url='test_url',
        headers={'foo': 'bar'},
        status_text='baz',
        status_code=200,
        raw_body=test_body,
    )
    conn.prep_response(response)

    assert response.method == 'get'
    assert response.url == 'test_url'
    assert response.headers == {'foo': 'bar'}
    assert response.status_code == 200
    assert response.status_text == 'baz'
    assert response.raw_body == test_body
    assert response.body == {'errorMessage': 'qux', 'errorNum': 1}
    assert response.error_code == 1
    assert response.error_message == 'qux'
    assert response.is_success is False



from __future__ import absolute_import, unicode_literals

import json

import pytest
# from requests.structures import CaseInsensitiveDict

from aioarangodb.exceptions import (
    ArangoServerError,
    DocumentInsertError,
    DocumentParseError,
    ArangoClientError
)
from aioarangodb.request import Request
from aioarangodb.response import Response
pytestmark = pytest.mark.asyncio

async def test_server_error(client, col, docs):
    document = docs[0]
    with pytest.raises(DocumentInsertError) as err:
        await col.insert(document, return_new=False)
        await col.insert(document, return_new=False)  # duplicate key error
    exc = err.value

    assert isinstance(exc, ArangoServerError)
    assert exc.source == 'server'
    assert exc.message == str(exc)
    assert exc.message.startswith('[HTTP 409][ERR 1210] unique constraint')
    assert exc.error_code == 1210
    assert exc.http_method == 'post'
    assert exc.http_code == 409
    # assert isinstance(exc.http_headers, CaseInsensitiveDict)

    resp = exc.response
    expected_body = {
        'code': exc.http_code,
        'error': True,
        'errorNum': exc.error_code,
        'errorMessage': exc.error_message
    }
    assert isinstance(resp, Response)
    assert resp.is_success is False
    assert resp.error_code == exc.error_code
    assert resp.body == expected_body
    assert resp.error_code == 1210
    assert resp.method == 'post'
    assert resp.status_code == 409
    assert resp.status_text == 'Conflict'

    assert json.loads(resp.raw_body) == expected_body
    assert resp.headers == exc.http_headers

    req = exc.request
    assert isinstance(req, Request)
    assert req.headers['content-type'] == 'application/json'
    assert req.method == 'post'
    assert req.params['silent'] == 0
    assert req.params['returnNew'] == 0
    assert req.data == document
    assert req.endpoint.startswith('/_api/document/' + col.name)


async def test_client_error(col):
    with pytest.raises(DocumentParseError) as err:
        await col.get({'_id': 'invalid'})  # malformed document
    exc = err.value

    assert isinstance(exc, ArangoClientError)
    assert exc.source == 'client'
    assert exc.error_code is None
    assert exc.error_message is None
    assert exc.message == str(exc)
    assert exc.message.startswith('bad collection name')
    assert exc.url is None
    assert exc.http_method is None
    assert exc.http_code is None
    assert exc.http_headers is None
    assert exc.response is None



from __future__ import absolute_import, unicode_literals

import json
import os

import pytest
from six import string_types

from aioarangodb.exceptions import (
    FoxxServiceGetError,
    FoxxServiceListError,
    FoxxServiceCreateError,
    FoxxServiceUpdateError,
    FoxxServiceReplaceError,
    FoxxServiceDeleteError,
    FoxxConfigGetError,
    FoxxConfigUpdateError,
    FoxxDependencyGetError,
    FoxxDependencyUpdateError,
    FoxxDependencyReplaceError,
    FoxxConfigReplaceError,
    FoxxDevModeEnableError,
    FoxxDevModeDisableError,
    FoxxReadmeGetError,
    FoxxSwaggerGetError,
    FoxxDownloadError,
    FoxxCommitError,
    FoxxScriptListError,
    FoxxScriptRunError,
    FoxxTestRunError
)
from aioarangodb.foxx import Foxx
from aioarangodb.tests.helpers import (
    assert_raises,
    extract,
    generate_service_mount
)
pytestmark = pytest.mark.asyncio


service_file = '/tmp/service.zip'


async def test_foxx_attributes(db):
    assert isinstance(db.foxx, Foxx)
    assert repr(db.foxx) == '<Foxx in {}>'.format(db.name)


async def test_foxx_service_management_json(db, bad_db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    service_mount = generate_service_mount()
    missing_mount = generate_service_mount()

    # Test list services
    for service in await db.foxx.services():
        assert 'development' in service
        assert 'legacy' in service
        assert 'mount' in service
        assert 'name' in service
        assert 'provides' in service
        assert 'version' in service

    # Test list services with bad database
    with assert_raises(FoxxServiceListError) as err:
        await bad_db.foxx.services()
    assert err.value.error_code in {11, 1228}

    # Test create service
    service = await db.foxx.create_service(
        mount=service_mount,
        source=service_file,
        config={},
        dependencies={},
        development=True,
        setup=True,
        legacy=True
    )
    assert service['mount'] == service_mount
    assert service['name'] == 'test'
    assert service['development'] is True
    assert service['legacy'] is True
    assert service['manifest']['configuration'] == {}
    assert service['manifest']['dependencies'] == {}

    # Test create duplicate service
    with assert_raises(FoxxServiceCreateError) as err:
        await db.foxx.create_service(service_mount, 'service.zip')
    assert err.value.error_code == 3011

    # Test get service
    service = await db.foxx.service(service_mount)
    assert service['mount'] == service_mount
    assert service['name'] == 'test'
    assert service['development'] is True
    assert service['manifest']['configuration'] == {}
    assert service['manifest']['dependencies'] == {}
    assert 'checksum' in service
    assert 'options' in service
    assert 'path' in service
    assert 'version' in service

    # Test get missing service
    with assert_raises(FoxxServiceGetError) as err:
        await db.foxx.service(missing_mount)
    assert err.value.error_code == 3009

    # Test update service
    service = await db.foxx.update_service(
        mount=service_mount,
        source=service_file,
        config={},
        dependencies={},
        teardown=True,
        setup=True,
        legacy=False,
        force=False
    )
    assert service['mount'] == service_mount
    assert service['name'] == 'test'
    assert service['legacy'] is False

    # Test update missing service
    with assert_raises(FoxxServiceUpdateError) as err:
        await db.foxx.update_service(missing_mount, 'service.zip')
    assert err.value.error_code == 3009

    # Test replace service
    service = await db.foxx.replace_service(
        mount=service_mount,
        source=service_file,
        config={},
        dependencies={},
        teardown=True,
        setup=True,
        legacy=True,
        force=False
    )
    assert service['mount'] == service_mount
    assert service['name'] == 'test'
    assert service['legacy'] is True

    # Test replace missing service
    with assert_raises(FoxxServiceReplaceError) as err:
        await db.foxx.replace_service(missing_mount, 'service.zip')
    assert err.value.error_code == 3009

    assert await db.foxx.delete_service(service_mount, teardown=False) is True
    assert service_mount not in await extract('mount', await db.foxx.services())

    # Test delete missing service
    with assert_raises(FoxxServiceDeleteError) as err:
        await db.foxx.delete_service(missing_mount, teardown=False)
    assert err.value.error_code == 3009


async def test_foxx_service_management_file(db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    path = os.path.join(os.path.dirname(__file__), 'static', 'service.zip')
    bad_path = os.path.join(os.path.dirname(__file__), 'static', 'service')

    service_mount = generate_service_mount()
    missing_mount = generate_service_mount()

    # Test create service by file with wrong extension
    with assert_raises(ValueError):
        await db.foxx.create_service_with_file(service_mount, bad_path)

    # Test create service by file
    service = await db.foxx.create_service_with_file(
        mount=service_mount,
        filename=path,
        development=True,
        setup=True,
        legacy=True
    )
    assert service['mount'] == service_mount
    assert service['name'] == 'test'
    assert service['development'] is True
    assert service['legacy'] is True
    assert service['manifest']['configuration'] == {}
    assert service['manifest']['dependencies'] == {}

    # Test create duplicate service
    with assert_raises(FoxxServiceCreateError) as err:
        await db.foxx.create_service_with_file(service_mount, path)
    assert err.value.error_code == 3011

    # Update config and dependencies
    assert await db.foxx.update_config(service_mount, {}) == {'values': {}}
    assert await db.foxx.update_dependencies(service_mount, {}) == {'values': {}}

    # Test update service by file
    service = await db.foxx.update_service_with_file(
        mount=service_mount,
        filename=path,
        teardown=False,
        setup=False,
        legacy=False,
        force=False
    )
    assert service['mount'] == service_mount
    assert service['name'] == 'test'
    assert service['legacy'] is False

    # Test update missing service
    with assert_raises(FoxxServiceUpdateError) as err:
        await db.foxx.update_service_with_file(missing_mount, path)
    assert err.value.error_code == 3009

    # Test replace service by file
    service = await db.foxx.replace_service_with_file(
        mount=service_mount,
        filename=path,
        teardown=True,
        setup=True,
        legacy=True,
        force=False
    )
    assert service['mount'] == service_mount
    assert service['name'] == 'test'
    assert service['legacy'] is True

    # Test replace missing service
    with assert_raises(FoxxServiceReplaceError) as err:
        await db.foxx.replace_service_with_file(missing_mount, path)
    assert err.value.error_code == 3009

    assert await db.foxx.delete_service(service_mount, teardown=False) is True
    assert service_mount not in await extract('mount', await db.foxx.services())


async def test_foxx_config_management(db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    service_mount = generate_service_mount()
    missing_mount = generate_service_mount()

    # Prep the test service
    await db.foxx.create_service(
        mount=service_mount,
        source=service_file,
        config={},
    )

    # Test get service config
    assert await db.foxx.config(service_mount) == {}

    # Test get missing service config
    with assert_raises(FoxxConfigGetError) as err:
        await db.foxx.config(missing_mount)
    assert err.value.error_code == 3009

    # Test update service config
    assert await db.foxx.update_config(service_mount, {}) == {'values': {}}

    # Test update missing service config
    with assert_raises(FoxxConfigUpdateError) as err:
        await db.foxx.update_config(missing_mount, {})
    assert err.value.error_code == 3009

    # Test replace service config
    assert await db.foxx.replace_config(service_mount, {}) == {'values': {}}

    # Test replace missing service config
    with assert_raises(FoxxConfigReplaceError) as err:
        await db.foxx.replace_config(missing_mount, {})
    assert err.value.error_code == 3009


async def test_foxx_dependency_management(db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    service_mount = generate_service_mount()
    missing_mount = generate_service_mount()

    # Prep the test service
    await db.foxx.create_service(
        mount=service_mount,
        source=service_file,
        dependencies={}
    )

    # Test get service dependencies
    assert await db.foxx.dependencies(service_mount) == {}

    # Test get missing service dependencies
    with assert_raises(FoxxDependencyGetError) as err:
        await db.foxx.dependencies(missing_mount)
    assert err.value.error_code == 3009

    # Test update service dependencies
    assert await db.foxx.update_dependencies(service_mount, {}) == {'values': {}}

    # Test update missing service dependencies
    with assert_raises(FoxxDependencyUpdateError) as err:
        await db.foxx.update_dependencies(missing_mount, {})
    assert err.value.error_code == 3009

    # Test replace service dependencies
    assert await db.foxx.replace_dependencies(service_mount, {}) == {'values': {}}

    # Test replace missing service dependencies
    with assert_raises(FoxxDependencyReplaceError) as err:
        await db.foxx.replace_dependencies(missing_mount, {})
    assert err.value.error_code == 3009


async def test_foxx_development_toggle(db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    service_mount = generate_service_mount()
    missing_mount = generate_service_mount()

    # Prep the test service
    await db.foxx.create_service(
        mount=service_mount,
        source=service_file,
        development=False,
    )

    # Test enable development mode
    service = await db.foxx.enable_development(service_mount)
    assert service['mount'] == service_mount
    assert service['name'] == 'test'
    assert service['development'] is True

    # Test enable development mode for missing service
    with assert_raises(FoxxDevModeEnableError) as err:
        await db.foxx.enable_development(missing_mount)
    assert err.value.error_code == 3009

    # Test disable development mode
    service = await db.foxx.disable_development(service_mount)
    assert service['mount'] == service_mount
    assert service['name'] == 'test'
    assert service['development'] is False

    # Test disable development mode for missing service
    with assert_raises(FoxxDevModeDisableError) as err:
        await db.foxx.disable_development(missing_mount)
    assert err.value.error_code == 3009


async def test_foxx_misc_functions(db, bad_db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    service_mount = generate_service_mount()
    missing_mount = generate_service_mount()

    # Prep the test service
    await db.foxx.create_service(
        mount=service_mount,
        source=service_file,
    )

    # Test get service readme
    assert 'Apache 2' in await db.foxx.readme(service_mount)

    # Test get missing service readme
    with assert_raises(FoxxReadmeGetError) as err:
        await db.foxx.readme(missing_mount)
    assert err.value.error_code == 3009

    # Test get service swagger
    swagger = await db.foxx.swagger(service_mount)
    assert 'swagger' in swagger
    assert 'paths' in swagger
    assert 'info' in swagger
    assert 'base_path' in swagger

    # Test get missing service swagger
    with assert_raises(FoxxSwaggerGetError) as err:
        await db.foxx.swagger(missing_mount)
    assert err.value.error_code == 3009

    # Test download service
    assert isinstance(await db.foxx.download(service_mount), string_types)

    # Test download missing service
    with assert_raises(FoxxDownloadError) as err:
        await db.foxx.download(missing_mount)
    assert err.value.error_code == 3009

    # Test commit service state
    assert await db.foxx.commit(replace=True) is True
    assert await db.foxx.commit(replace=False) is True

    # Test commit service state with bad database
    with assert_raises(FoxxCommitError) as err:
        await bad_db.foxx.commit(replace=True)
    assert err.value.error_code in {11, 1228}

    # Test list service scripts
    scripts = await db.foxx.scripts(service_mount)
    assert 'setup' in scripts
    assert 'teardown' in scripts

    # Test list missing service scripts
    with assert_raises(FoxxScriptListError) as err:
        await db.foxx.scripts(missing_mount)
    assert err.value.error_code == 3009

    # Test run service script
    assert await db.foxx.run_script(service_mount, 'setup', []) == {}
    assert await db.foxx.run_script(service_mount, 'teardown', []) == {}

    # Test run missing service script
    with assert_raises(FoxxScriptRunError) as err:
        await db.foxx.run_script(service_mount, 'invalid', ())
    assert err.value.error_code == 3016

    # Test run tests on service
    result_str = await db.foxx.run_tests(
        mount=service_mount,
        reporter='suite',
        idiomatic=True,
        name_filter='science'
    )
    result_json = json.loads(result_str)
    assert 'stats' in result_json
    assert 'tests' in result_json
    assert 'suites' in result_json

    result_str = await db.foxx.run_tests(
        mount=service_mount,
        reporter='stream',
        output_format='x-ldjson'
    )
    for result_part in result_str.split('\r\n'):
        if len(result_part) == 0:
            continue
        assert result_part.startswith('[')
        assert result_part.endswith(']')

    result_str = db.foxx.run_tests(
        mount=service_mount,
        reporter='stream',
        output_format='text'
    )
    assert result_str.startswith('[')
    assert result_str.endswith(']') or result_str.endswith('\r\n')

    result_str = await db.foxx.run_tests(
        mount=service_mount,
        reporter='xunit',
        output_format='xml'
    )
    assert result_str.strip().startswith('<')
    assert result_str.strip().endswith('>')

    # Test run tests on missing service
    with assert_raises(FoxxTestRunError) as err:
        await db.foxx.run_tests(missing_mount)
    assert err.value.error_code == 3009



from __future__ import absolute_import, unicode_literals

__all__ = [
    'DefaultExecutor',
    'AsyncExecutor',
    'BatchExecutor',
    'TransactionExecutor'
]

from collections import OrderedDict
from uuid import uuid4

from six import moves

from .exceptions import (
    AsyncExecuteError,
    BatchStateError,
    BatchExecuteError,
    TransactionAbortError,
    TransactionCommitError,
    TransactionStatusError,
    TransactionInitError
)
from .job import (
    AsyncJob,
    BatchJob
)
from .request import Request
from .response import Response
from .utils import suppress_warning


class Executor(object):  # pragma: no cover
    """Base class for API executors.

    API executors dictate how API requests are executed depending on the
    execution context (i.e. "default", "async", "batch", "transaction").

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    """
    context = None

    def __init__(self, connection):
        self._conn = connection

    async def execute(self, request, response_handler):
        """Execute an API request.

        :param request: HTTP request.
        :type request: arango.request.Request
        :param response_handler: HTTP response handler.
        :type response_handler: callable
        :return: API execution result or job.
        :rtype: str | unicode | bool | int | list | dict | arango.job.Job
        """
        raise NotImplementedError


class DefaultExecutor(Executor):
    """Default API executor.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    """
    context = 'default'

    def __init__(self, connection):
        super(DefaultExecutor, self).__init__(connection)

    async def execute(self, request, response_handler):
        """Execute an API request and return the result.

        :param request: HTTP request.
        :type request: arango.request.Request
        :param response_handler: HTTP response handler.
        :type response_handler: callable
        :return: API execution result.
        :rtype: str | unicode | bool | int | list | dict
        """
        resp = await self._conn.send_request(request)
        return response_handler(resp)


class AsyncExecutor(Executor):
    """Async API Executor.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param return_result: If set to True, API executions return instances of
        :class:`arango.job.AsyncJob` and results can be retrieved from server
        once available. If set to False, API executions return None and no
        results are stored on server.
    :type return_result: bool
    """
    context = 'async'

    def __init__(self, connection, return_result):
        super(AsyncExecutor, self).__init__(connection)
        self._return_result = return_result

    async def execute(self, request, response_handler):
        """Execute an API request asynchronously.

        :param request: HTTP request.
        :type request: arango.request.Request
        :param response_handler: HTTP response handler.
        :type response_handler: callable
        :return: Async job or None if **return_result** parameter was set to
            False during initialization.
        :rtype: arango.job.AsyncJob | None
        """
        if self._return_result:
            request.headers['x-arango-async'] = 'store'
        else:
            request.headers['x-arango-async'] = 'true'

        resp = await self._conn.send_request(request)
        if not resp.is_success:
            raise AsyncExecuteError(resp, request)
        if not self._return_result:
            return None

        job_id = resp.headers['x-arango-async-id']
        return AsyncJob(self._conn, job_id, response_handler)


class BatchExecutor(Executor):
    """Batch API executor.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param return_result: If set to True, API executions return instances of
        :class:`arango.job.BatchJob` that are populated with results on commit.
        If set to False, API executions return None and no results are tracked
        client-side.
    :type return_result: bool
    """
    context = 'batch'

    def __init__(self, connection, return_result):
        super(BatchExecutor, self).__init__(connection)
        self._return_result = return_result
        self._queue = OrderedDict()
        self._committed = False

    # noinspection PyMethodMayBeStatic,PyUnresolvedReferences
    def _stringify_request(self, request):
        path = request.endpoint

        if request.params is not None:
            path += '?' + moves.urllib.parse.urlencode(request.params)
        buffer = ['{} {} HTTP/1.1'.format(request.method, path)]

        if request.headers is not None:
            for key, value in sorted(request.headers.items()):
                buffer.append('{}: {}'.format(key, value))

        if request.data is not None:
            serialized = self._conn.serialize(request.data)
            buffer.append('\r\n' + serialized)

        return '\r\n'.join(buffer)

    @property
    def jobs(self):
        """Return the queued batch jobs.

        :return: Batch jobs or None if **return_result** parameter was set to
            False during initialization.
        :rtype: [arango.job.BatchJob] | None
        """
        if not self._return_result:
            return None
        return [job for _, job in self._queue.values()]

    async def execute(self, request, response_handler):
        """Place the request in the batch queue.

        :param request: HTTP request.
        :type request: arango.request.Request
        :param response_handler: HTTP response handler.
        :type response_handler: callable
        :return: Batch job or None if **return_result** parameter was set to
            False during initialization.
        :rtype: arango.job.BatchJob | None
        :raise arango.exceptions.BatchStateError: If batch was already
            committed.
        """
        if self._committed:
            raise BatchStateError('batch already committed')

        job = BatchJob(response_handler)
        self._queue[job.id] = (request, job)
        return job if self._return_result else None

    async def commit(self):
        """Execute the queued requests in a single batch API request.

        If **return_result** parameter was set to True during initialization,
        :class:`arango.job.BatchJob` instances are populated with results.

        :return: Batch jobs or None if **return_result** parameter was set to
            False during initialization.
        :rtype: [arango.job.BatchJob] | None
        :raise arango.exceptions.BatchStateError: If batch state is invalid
            (e.g. batch was already committed or size of response from server
            did not match the expected).
        :raise arango.exceptions.BatchExecuteError: If commit fails.
        """
        if self._committed:
            raise BatchStateError('batch already committed')
        else:
            self._committed = True

        if len(self._queue) == 0:
            return self.jobs

        # Boundary used for multipart request
        boundary = uuid4().hex

        # Build the batch request payload
        buffer = []
        for req, job in self._queue.values():
            buffer.append('--{}'.format(boundary))
            buffer.append('Content-Type: application/x-arango-batchpart')
            buffer.append('Content-Id: {}'.format(job.id))
            buffer.append('\r\n' + self._stringify_request(req))
        buffer.append('--{}--'.format(boundary))

        request = Request(
            method='post',
            endpoint='/_api/batch',
            headers={
                'Content-Type':
                    'multipart/form-data; boundary={}'.format(boundary)
            },
            data='\r\n'.join(buffer),
        )
        with suppress_warning('requests.packages.urllib3.connectionpool'):
            resp = await self._conn.send_request(request)

        if not resp.is_success:
            raise BatchExecuteError(resp, request)

        if not self._return_result:
            return None

        url_prefix = resp.url.strip('/_api/batch')
        raw_resps = resp.raw_body.split('--{}'.format(boundary))[1:-1]

        if len(self._queue) != len(raw_resps):
            raise BatchStateError(
                'expecting {} parts in batch response but got {}'
                .format(len(self._queue), len(raw_resps))
            )
        for raw_resp in raw_resps:
            # Parse and breakdown the batch response body
            resp_parts = raw_resp.strip().split('\r\n')
            raw_content_id = resp_parts[1]
            raw_body = resp_parts[-1]
            raw_status = resp_parts[3]
            job_id = raw_content_id.split(' ')[1]
            _, status_code, status_text = raw_status.split(' ', 2)

            # Update the corresponding batch job
            queued_req, queued_job = self._queue[job_id]

            queued_job._status = 'done'
            resp = Response(
                method=queued_req.method,
                url=url_prefix + queued_req.endpoint,
                headers={},
                status_code=int(status_code),
                status_text=status_text,
                raw_body=raw_body
            )
            queued_job._response = self._conn.prep_response(resp)

        return self.jobs


class TransactionExecutor(Executor):
    """Executes transaction API requests.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param read: Name(s) of collections read during transaction. Read-only
        collections are added lazily but should be declared if possible to
        avoid deadlocks.
    :type read: str | unicode | [str | unicode]
    :param write: Name(s) of collections written to during transaction with
        shared access.
    :type write: str | unicode | [str | unicode]
    :param exclusive: Name(s) of collections written to during transaction
        with exclusive access.
    :type exclusive: str | unicode | [str | unicode]
    :param sync: Block until operation is synchronized to disk.
    :type sync: bool
    :param allow_implicit: Allow reading from undeclared collections.
    :type allow_implicit: bool
    :param lock_timeout: Timeout for waiting on collection locks. If not given,
        a default value is used. Setting it to 0 disables the timeout.
    :type lock_timeout: int
    :param max_size: Max transaction size in bytes. Applicable to RocksDB
        storage engine only.
    :type max_size: int
    """
    context = 'transaction'

    async def begin(
            self,
            read=None,
            write=None,
            exclusive=None,
            sync=None,
            allow_implicit=None,
            lock_timeout=None,
            max_size=None):
        collections = {}
        if read is not None:
            collections['read'] = read
        if write is not None:
            collections['write'] = write
        if exclusive is not None:
            collections['exclusive'] = exclusive

        data = {'collections': collections}
        if sync is not None:
            data['waitForSync'] = sync
        if allow_implicit is not None:
            data['allowImplicit'] = allow_implicit
        if lock_timeout is not None:
            data['lockTimeout'] = lock_timeout
        if max_size is not None:
            data['maxTransactionSize'] = max_size

        request = Request(
            method='post',
            endpoint='/_api/transaction/begin',
            data=data
        )
        resp = await self._conn.send_request(request)

        if not resp.is_success:
            raise TransactionInitError(resp, request)

        self._id = resp.body['result']['id']

    @property
    def id(self):
        """Return the transaction ID.

        :return: Transaction ID.
        :rtype: str | unicode
        """
        return self._id

    async def execute(self, request, response_handler):
        """Execute API request in a transaction and return the result.

        :param request: HTTP request.
        :type request: arango.request.Request
        :param response_handler: HTTP response handler.
        :type response_handler: callable
        :return: API execution result.
        :rtype: str | unicode | bool | int | list | dict
        """
        request.headers['x-arango-trx-id'] = self._id
        resp = await self._conn.send_request(request)
        return response_handler(resp)

    async def status(self):
        """Return the transaction status.

        :return: Transaction status.
        :rtype: str | unicode
        :raise arango.exceptions.TransactionStatusError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/transaction/{}'.format(self._id),
        )
        resp = await self._conn.send_request(request)

        if resp.is_success:
            return resp.body['result']['status']
        raise TransactionStatusError(resp, request)

    async def commit(self):
        """Commit the transaction.

        :return: True if commit was successful.
        :rtype: bool
        :raise arango.exceptions.TransactionCommitError: If commit fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/transaction/{}'.format(self._id),
        )
        resp = await self._conn.send_request(request)

        if resp.is_success:
            return True
        raise TransactionCommitError(resp, request)

    async def abort(self):
        """Abort the transaction.

        :return: True if the abort operation was successful.
        :rtype: bool
        :raise arango.exceptions.TransactionAbortError: If abort fails.
        """
        request = Request(
            method='delete',
            endpoint='/_api/transaction/{}'.format(self._id),
        )
        resp = await self._conn.send_request(request)

        if resp.is_success:
            return True
        raise TransactionAbortError(resp, request)



from __future__ import absolute_import, unicode_literals

import pytest

from aioarangodb.errno import (
    DATABASE_NOT_FOUND,
    FORBIDDEN
)
from aioarangodb.exceptions import (
    ClusterEndpointsError,
    ClusterHealthError,
    ClusterMaintenanceModeError,
    ClusterServerEngineError,
    ClusterServerIDError,
    ClusterServerRoleError,
    ClusterServerStatisticsError,
    ClusterServerVersionError,
)
from aioarangodb.tests.helpers import assert_raises
pytestmark = pytest.mark.asyncio


async def test_cluster_server_id(sys_db, bad_db, cluster):
    if not cluster:
        pytest.skip('Only tested in a cluster setup')

    result = await sys_db.cluster.server_id()
    assert isinstance(result, str)

    with assert_raises(ClusterServerIDError) as err:
        await bad_db.cluster.server_id()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_cluster_server_role(sys_db, bad_db, cluster):
    if not cluster:
        pytest.skip('Only tested in a cluster setup')

    result = await sys_db.cluster.server_role()
    assert isinstance(result, str)

    with assert_raises(ClusterServerRoleError) as err:
        await bad_db.cluster.server_role()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_cluster_health(sys_db, bad_db, cluster):
    if not cluster:
        pytest.skip('Only tested in a cluster setup')

    result = await sys_db.cluster.health()
    assert 'Health' in result
    assert 'ClusterId' in result

    with assert_raises(ClusterHealthError) as err:
        await bad_db.cluster.health()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_cluster_server_version(sys_db, bad_db, cluster):
    if not cluster:
        pytest.skip('Only tested in a cluster setup')

    server_id = await sys_db.cluster.server_id()
    result = await sys_db.cluster.server_version(server_id)
    assert 'server' in result
    assert 'version' in result

    with assert_raises(ClusterServerVersionError) as err:
        await bad_db.cluster.server_version(server_id)
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_cluster_server_engine(sys_db, bad_db, cluster):
    if not cluster:
        pytest.skip('Only tested in a cluster setup')

    server_id = await sys_db.cluster.server_id()
    result = await sys_db.cluster.server_engine(server_id)
    assert 'name' in result
    assert 'supports' in result

    with assert_raises(ClusterServerEngineError) as err:
        bad_db.cluster.server_engine(server_id)
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_cluster_server_statistics(sys_db, bad_db, cluster):
    if not cluster:
        pytest.skip('Only tested in a cluster setup')

    server_id = await sys_db.cluster.server_id()
    result = await sys_db.cluster.server_statistics(server_id)
    assert 'time' in result
    assert 'system' in result
    assert 'enabled' in result

    with assert_raises(ClusterServerStatisticsError) as err:
        await bad_db.cluster.server_statistics(server_id)
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_cluster_toggle_maintenance_mode(sys_db, bad_db, cluster):
    if not cluster:
        pytest.skip('Only tested in a cluster setup')

    result = await sys_db.cluster.toggle_maintenance_mode('on')
    assert 'error' in result
    assert 'warning' in result

    result = await sys_db.cluster.toggle_maintenance_mode('off')
    assert 'error' in result
    assert 'warning' in result

    with assert_raises(ClusterMaintenanceModeError) as err:
        await bad_db.cluster.toggle_maintenance_mode('on')
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_cluster_endpoints(db, bad_db, cluster):
    if not cluster:
        pytest.skip('Only tested in a cluster setup')

    # Test get server endpoints
    assert len(await db.cluster.endpoints()) > 0

    # Test get server endpoints with bad database
    with assert_raises(ClusterEndpointsError) as err:
        await bad_db.cluster.endpoints()
    assert err.value.error_code in {11, 1228}



from aioarangodb.tests import global_data
from aioarangodb.tests.helpers import generate_db_name
pytest_plugins = ['aioarangodb.tests.fixtures']


def pytest_addoption(parser):
    parser.addoption('--passwd', action='store', default='secret')
    parser.addoption('--complete', action='store_true')
    parser.addoption('--cluster', action='store_true')
    parser.addoption('--replication', action='store_true')
    parser.addoption('--secret', action='store', default='secret')
    parser.addoption('--tst_db_name', action='store', default=generate_db_name())

def pytest_configure(config):
    global_data['secret'] = config.getoption('secret')
    global_data['tst_db_name'] = config.getoption('tst_db_name')
    global_data['replication'] = config.getoption('replication')
    global_data['cluster'] = config.getoption('cluster')
    global_data['complete'] = config.getoption('complete')
    global_data['passwd'] = config.getoption('passwd')
    global_data['username'] = 'root'



from __future__ import absolute_import, unicode_literals
import pytest
from aioarangodb.exceptions import (
    AQLCacheClearError,
    AQLCacheConfigureError,
    AQLCacheEntriesError,
    AQLCachePropertiesError,
    AQLFunctionCreateError,
    AQLFunctionDeleteError,
    AQLFunctionListError,
    AQLQueryClearError,
    AQLQueryExecuteError,
    AQLQueryExplainError,
    AQLQueryListError,
    AQLQueryTrackingGetError,
    AQLQueryTrackingSetError,
    AQLQueryKillError,
    AQLQueryValidateError
)
from aioarangodb.tests.helpers import assert_raises, extract
pytestmark = pytest.mark.asyncio


async def test_aql_attributes(db, username):
    assert db.context in ['default', 'async', 'batch', 'transaction']
    assert db.username == username
    assert db.db_name == db.name
    assert repr(db.aql) == '<AQL in {}>'.format(db.name)
    assert repr(db.aql.cache) == '<AQLQueryCache in {}>'.format(db.name)


async def test_aql_query_management(db, bad_db, col, docs):
    plan_fields = [
        'estimatedNrItems',
        'estimatedCost',
        'rules',
        'variables',
        'collections',
    ]
    # Test explain invalid query
    with assert_raises(AQLQueryExplainError) as err:
        await db.aql.explain('INVALID QUERY')
    assert err.value.error_code == 1501

    # Test explain valid query with all_plans set to False
    plan = await db.aql.explain(
        'FOR d IN {} RETURN d'.format(col.name),
        all_plans=False,
        opt_rules=['-all', '+use-index-range']
    )
    assert all(field in plan for field in plan_fields)

    # Test explain valid query with all_plans set to True
    plans = await db.aql.explain(
        'FOR d IN {} RETURN d'.format(col.name),
        all_plans=True,
        opt_rules=['-all', '+use-index-range'],
        max_plans=10
    )
    for plan in plans:
        assert all(field in plan for field in plan_fields)
    assert len(plans) < 10

    # Test validate invalid query
    with assert_raises(AQLQueryValidateError) as err:
        await db.aql.validate('INVALID QUERY')
    assert err.value.error_code == 1501

    # Test validate valid query
    result = await db.aql.validate('FOR d IN {} RETURN d'.format(col.name))
    assert 'ast' in result
    assert 'bind_vars' in result
    assert 'collections' in result
    assert 'parsed' in result

    # Test execute invalid AQL query
    with assert_raises(AQLQueryExecuteError) as err:
        await db.aql.execute('INVALID QUERY')
    assert err.value.error_code == 1501

    # Test execute valid query
    await db.collection(col.name).import_bulk(docs)
    cursor = await db.aql.execute(
        '''
        FOR d IN {col}
            UPDATE {{_key: d._key, _val: @val }} IN {col}
            RETURN NEW
        '''.format(col=col.name),
        count=True,
        # batch_size=1,
        ttl=10,
        bind_vars={'val': 42},
        full_count=True,
        max_plans=1000,
        optimizer_rules=['+all'],
        cache=True,
        memory_limit=1000000,
        fail_on_warning=False,
        profile=True,
        max_transaction_size=100000,
        max_warning_count=10,
        intermediate_commit_count=1,
        intermediate_commit_size=1000,
        satellite_sync_wait=False,
        write_collections=[col.name],
        read_collections=[col.name],
        stream=False,
        skip_inaccessible_cols=True,
        max_runtime=0.0
    )
    assert cursor.id is None
    assert cursor.type == 'cursor'
    assert cursor.batch() is not None
    assert cursor.has_more() is False
    assert cursor.count() == await col.count()
    assert cursor.cached() is False
    assert cursor.statistics() is not None
    assert cursor.profile() is not None
    assert cursor.warnings() == []
    assert await extract('_key', cursor) == await extract('_key', docs)
    assert await cursor.close(ignore_missing=True) is None

    # Test get tracking properties with bad database
    with assert_raises(AQLQueryTrackingGetError) as err:
        await bad_db.aql.tracking()
    assert err.value.error_code in {11, 1228}

    # Test get tracking properties
    tracking = await db.aql.tracking()
    assert isinstance(tracking['enabled'], bool)
    assert isinstance(tracking['max_query_string_length'], int)
    assert isinstance(tracking['max_slow_queries'], int)
    assert isinstance(tracking['slow_query_threshold'], int)
    assert isinstance(tracking['track_bind_vars'], bool)
    assert isinstance(tracking['track_slow_queries'], bool)

    # Test set tracking properties with bad database
    with assert_raises(AQLQueryTrackingSetError) as err:
        await bad_db.aql.set_tracking(enabled=not tracking['enabled'])
    assert err.value.error_code in {11, 1228}
    assert (await db.aql.tracking())['enabled'] == tracking['enabled']

    # Test set tracking properties
    new_tracking = await db.aql.set_tracking(
        enabled=not tracking['enabled'],
        max_query_string_length=4000,
        max_slow_queries=60,
        slow_query_threshold=15,
        track_bind_vars=not tracking['track_bind_vars'],
        track_slow_queries=not tracking['track_slow_queries']
    )
    assert new_tracking['enabled'] != tracking['enabled']
    assert new_tracking['max_query_string_length'] == 4000
    assert new_tracking['max_slow_queries'] == 60
    assert new_tracking['slow_query_threshold'] == 15
    assert new_tracking['track_bind_vars'] != tracking['track_bind_vars']
    assert new_tracking['track_slow_queries'] != tracking['track_slow_queries']

    # Make sure to revert the properties
    new_tracking = await db.aql.set_tracking(
        enabled=True,
        track_bind_vars=True,
        track_slow_queries=True
    )
    assert new_tracking['enabled'] is True
    assert new_tracking['track_bind_vars'] is True
    assert new_tracking['track_slow_queries'] is True

    # Kick off some long lasting queries in the background
    await db.begin_async_execution().aql.execute('RETURN SLEEP(100)')
    await db.begin_async_execution().aql.execute('RETURN SLEEP(50)')

    # Test list queries
    queries = await db.aql.queries()
    for query in queries:
        assert 'id' in query
        assert 'query' in query
        assert 'started' in query
        assert 'state' in query
        assert 'bind_vars' in query
        assert 'runtime' in query
    assert len(queries) == 2

    # Test list queries with bad database
    with assert_raises(AQLQueryListError) as err:
        await bad_db.aql.queries()
    assert err.value.error_code in {11, 1228}

    # Test kill queries
    query_id_1, query_id_2 = await extract('id', queries)
    assert await db.aql.kill(query_id_1) is True

    while len(queries) > 1:
        queries = await db.aql.queries()
    assert query_id_1 not in await extract('id', queries)

    assert await db.aql.kill(query_id_2) is True
    while len(queries) > 0:
        queries = await db.aql.queries()
    assert query_id_2 not in await extract('id', queries)

    # Test kill missing queries
    with assert_raises(AQLQueryKillError) as err:
        await db.aql.kill(query_id_1)
    assert err.value.error_code == 1591
    with assert_raises(AQLQueryKillError) as err:
        await db.aql.kill(query_id_2)
    assert err.value.error_code == 1591

    # Test list slow queries
    assert await db.aql.slow_queries() == []

    # Test list slow queries with bad database
    with assert_raises(AQLQueryListError) as err:
        await bad_db.aql.slow_queries()
    assert err.value.error_code in {11, 1228}

    # Test clear slow queries
    assert await db.aql.clear_slow_queries() is True

    # Test clear slow queries with bad database
    with assert_raises(AQLQueryClearError) as err:
        await bad_db.aql.clear_slow_queries()
    assert err.value.error_code in {11, 1228}


async def test_aql_function_management(db, bad_db):
    fn_group = 'functions::temperature'
    fn_name_1 = 'functions::temperature::celsius_to_fahrenheit'
    fn_body_1 = 'function (celsius) { return celsius * 1.8 + 32; }'
    fn_name_2 = 'functions::temperature::fahrenheit_to_celsius'
    fn_body_2 = 'function (fahrenheit) { return (fahrenheit - 32) / 1.8; }'
    bad_fn_name = 'functions::temperature::should_not_exist'
    bad_fn_body = 'function (celsius) { invalid syntax }'

    # Test list AQL functions
    assert await db.aql.functions() == []

    # Test list AQL functions with bad database
    with assert_raises(AQLFunctionListError) as err:
        await bad_db.aql.functions()
    assert err.value.error_code in {11, 1228}

    # Test create invalid AQL function
    with assert_raises(AQLFunctionCreateError) as err:
        await db.aql.create_function(bad_fn_name, bad_fn_body)
    assert err.value.error_code == 1581

    # Test create AQL function one
    assert await db.aql.create_function(fn_name_1, fn_body_1) == {'is_new': True}
    functions = await db.aql.functions()
    assert len(functions) == 1
    assert functions[0]['name'] == fn_name_1
    assert functions[0]['code'] == fn_body_1
    assert 'is_deterministic' in functions[0]

    # Test create AQL function one again (idempotency)
    assert await db.aql.create_function(fn_name_1, fn_body_1) == {'is_new': False}
    functions = await db.aql.functions()
    assert len(functions) == 1
    assert functions[0]['name'] == fn_name_1
    assert functions[0]['code'] == fn_body_1
    assert 'is_deterministic' in functions[0]

    # Test create AQL function two
    assert await db.aql.create_function(fn_name_2, fn_body_2) == {'is_new': True}
    functions = sorted(await db.aql.functions(), key=lambda x: x['name'])
    assert len(functions) == 2
    assert functions[0]['name'] == fn_name_1
    assert functions[0]['code'] == fn_body_1
    assert functions[1]['name'] == fn_name_2
    assert functions[1]['code'] == fn_body_2
    assert 'is_deterministic' in functions[0]
    assert 'is_deterministic' in functions[1]

    # Test delete AQL function one
    assert await db.aql.delete_function(fn_name_1) == {'deleted': 1}
    functions = await db.aql.functions()
    assert len(functions) == 1
    assert functions[0]['name'] == fn_name_2
    assert functions[0]['code'] == fn_body_2

    # Test delete missing AQL function
    with assert_raises(AQLFunctionDeleteError) as err:
        await db.aql.delete_function(fn_name_1)
    assert err.value.error_code == 1582
    assert await db.aql.delete_function(fn_name_1, ignore_missing=True) is False
    functions = await db.aql.functions()
    assert len(functions) == 1
    assert functions[0]['name'] == fn_name_2
    assert functions[0]['code'] == fn_body_2

    # Test delete AQL function group
    assert await db.aql.delete_function(fn_group, group=True) == {'deleted': 1}
    assert await db.aql.functions() == []


async def test_aql_cache_management(db, bad_db):
    # Test get AQL cache properties
    properties = await db.aql.cache.properties()
    assert 'mode' in properties
    assert 'max_results' in properties
    assert 'max_results_size' in properties
    assert 'max_entry_size' in properties
    assert 'include_system' in properties

    # Test get AQL cache properties with bad database
    with assert_raises(AQLCachePropertiesError):
        await bad_db.aql.cache.properties()

    # Test get AQL cache configure properties
    properties = await db.aql.cache.configure(
        mode='on',
        max_results=100,
        max_results_size=10000,
        max_entry_size=10000,
        include_system=True
    )
    assert properties['mode'] == 'on'
    assert properties['max_results'] == 100
    assert properties['max_results_size'] == 10000
    assert properties['max_entry_size'] == 10000
    assert properties['include_system'] is True

    properties = await db.aql.cache.properties()
    assert properties['mode'] == 'on'
    assert properties['max_results'] == 100
    assert properties['max_results_size'] == 10000
    assert properties['max_entry_size'] == 10000
    assert properties['include_system'] is True

    # Test get AQL cache configure properties with bad database
    with assert_raises(AQLCacheConfigureError):
        await bad_db.aql.cache.configure(mode='on')

    # Test get AQL cache entries
    result = await db.aql.cache.entries()
    assert isinstance(result, list)

    # Test get AQL cache entries with bad database
    with assert_raises(AQLCacheEntriesError) as err:
        await bad_db.aql.cache.entries()
    assert err.value.error_code in {11, 1228}

    # Test get AQL cache clear
    result = await db.aql.cache.clear()
    assert isinstance(result, bool)

    # Test get AQL cache clear with bad database
    with assert_raises(AQLCacheClearError) as err:
        await bad_db.aql.cache.clear()
    assert err.value.error_code in {11, 1228}
